# Copyright (c) Acconeer AB, 2025-2026
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, List, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtGui import QFont
from PySide6.QtWidgets import QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo.distance._detector import ConfigMismatchError
from acconeer.exptool.a121.algo.distance._translation import detector_config_to_session_config
from acconeer.exptool.a121.algo.presence._detector import Detector as PresenceDetector
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import CollapsibleWidget
from acconeer.exptool.app.new.ui.components.a121 import (
    SensorConfigEditor,
)

from ._configs import (
    get_10_ft_container_config,
    get_20_ft_container_config,
    get_40_ft_container_config,
    get_no_lens_config,
)
from ._ex_app import (
    PRESENCE_RUN_TIME_S,
    CargoPresenceConfig,
    ContainerSize,
    ExApp,
    ExAppConfig,
    ExAppContext,
    ExAppResult,
    UtilizationLevelConfig,
    _load_algo_data,
    _Mode,
)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: ExAppConfig = attrs.field(factory=get_20_ft_container_config)
    ex_app_context: Optional[ExAppContext] = attrs.field(default=None)


class PluginPresetId(Enum):
    CONTAINER_10_FT = auto()
    CONTAINER_20_FT = auto()
    CONTAINER_40_FT = auto()
    NO_LENS = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    ex_app_config: ExAppConfig
    estimated_frame_rate: Optional[float]
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], ExAppConfig]] = {
        PluginPresetId.CONTAINER_10_FT.value: lambda: get_10_ft_container_config(),
        PluginPresetId.CONTAINER_20_FT.value: lambda: get_20_ft_container_config(),
        PluginPresetId.CONTAINER_40_FT.value: lambda: get_40_ft_container_config(),
        PluginPresetId.NO_LENS.value: lambda: get_no_lens_config(),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key, use_app_client=False)

        self._recorder: Optional[a121.H5Recorder] = None
        self._ex_app_instance: Optional[ExApp] = None
        self._log = BackendLogger.getLogger(__name__)
        self._frame_count = 0

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = ExAppConfig.from_json(file["config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState(config=get_no_lens_config())
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: ExAppConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_utilization_level_config(self, *, config: UtilizationLevelConfig) -> None:
        self.shared_state.config.utilization_level_config = config
        self.broadcast()

    @is_task
    def update_cargo_presence_config(self, *, config: CargoPresenceConfig) -> None:
        self.shared_state.config.cargo_presence_config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, ex_app_context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.ex_app_context = ex_app_context
        self.shared_state.sensor_id = record.session(0).sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client

        if (
            self.shared_state.config.cargo_presence_config is not None
            and self.shared_state.ex_app_context is not None
            and self.shared_state.ex_app_context.presence_context is not None
        ):
            update_rate_diff = abs(
                self.shared_state.ex_app_context.presence_context.estimated_frame_rate
                - self.shared_state.config.cargo_presence_config.update_rate
            )
            if update_rate_diff > 0.01:
                self.shared_state.ex_app_context.presence_context = None
                self.send_status_message("Config mismatch, estimating frame rate.")

        try:
            self._ex_app_instance = ExApp(
                client=self.client,
                sensor_id=self.shared_state.sensor_id,
                ex_app_config=self.shared_state.config,
                ex_app_context=self.shared_state.ex_app_context,
            )
            self._ex_app_instance.start(recorder)
        except ConfigMismatchError:
            self._ex_app_instance = ExApp(
                client=self.client,
                sensor_id=self.shared_state.sensor_id,
                ex_app_config=self.shared_state.config,
                ex_app_context=None,
            )
            self.send_status_message("Config mismatch, recalibrating detector.")
            self._ex_app_instance.start(recorder)

        self.shared_state.ex_app_context = self._ex_app_instance.ex_app_context

        if self._ex_app_instance.presence_context is not None:
            estimated_frame_rate = self._ex_app_instance.presence_context.estimated_frame_rate
        else:
            estimated_frame_rate = None

        self.callback(
            SetupMessage(
                ex_app_config=self.shared_state.config,
                estimated_frame_rate=estimated_frame_rate,
            )
        )

    def end_session(self) -> None:
        if self._ex_app_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._ex_app_instance.stop()
        self._frame_count = 0

    def get_next(self) -> None:
        assert self.client
        if self._ex_app_instance is None:
            raise RuntimeError
        result = self._ex_app_instance.get_next()
        self._frame_count += 1

        self.callback(backend.PlotMessage(result=result))
        self.callback(GeneralMessage(name="frame_count", data=self._frame_count))


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_jobs: List[ExAppResult] = []
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_jobs.append(message.result)
        elif isinstance(message, SetupMessage):
            self.setup(
                message.ex_app_config,
                message.estimated_frame_rate,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_jobs is None:
            return

        plot_distance = True
        plot_presence = True
        try:
            for plot_job in self._plot_jobs:
                if (plot_job.mode == _Mode.DISTANCE and plot_distance) or (
                    plot_job.mode == _Mode.PRESENCE and plot_presence
                ):
                    self.draw_plot_job(result=plot_job)
        finally:
            self._plot_jobs = []

    def setup(
        self,
        ex_app_config: ExAppConfig,
        estimated_frame_rate: Optional[float],
    ) -> None:
        self.plot_layout.clear()

        self.ex_app_config = ex_app_config
        self.utilization_level_config = ex_app_config.utilization_level_config
        self.cargo_presence_config = ex_app_config.cargo_presence_config

        win = self.plot_layout
        self.num_rects = 16

        # Utilization level

        if self.ex_app_config.activate_utilization_level:
            # Distance sweep plot

            self.sweep_plot = win.addPlot(row=0, col=0, title="Distance sweep")
            self.sweep_plot.setMenuEnabled(False)
            self.sweep_plot.showGrid(x=True, y=True)
            self.sweep_plot.addLegend()
            self.sweep_plot.setLabel("left", "Amplitude")
            self.sweep_plot.setLabel("bottom", "Distance (m)")
            self.sweep_plot.addItem(pg.PlotDataItem())

            self.num_curves = 4

            pen = et.utils.pg_pen_cycler(0)
            brush = et.utils.pg_brush_cycler(0)
            symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
            feat_kw = dict(pen=pen, **symbol_kw)
            self.sweep_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

            pen = et.utils.pg_pen_cycler(1)
            brush = et.utils.pg_brush_cycler(1)
            symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
            feat_kw = dict(pen=pen, **symbol_kw)
            self.threshold_curves = [
                self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)
            ]

            sweep_plot_legend = pg.LegendItem(offset=(-30, 30))
            sweep_plot_legend.setParentItem(self.sweep_plot.graphicsItem())
            sweep_plot_legend.addItem(self.sweep_curves[0], "Sweep")
            sweep_plot_legend.addItem(self.threshold_curves[0], "Threshold")

            font = QFont()
            font.setPixelSize(16)
            self.sweep_text_item = pg.TextItem(
                fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
                anchor=(0.5, 0),
                color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
            )
            self.sweep_text_item.setFont(font)
            self.sweep_text_item.hide()
            self.sweep_plot.addItem(self.sweep_text_item)

            self.sweep_main_peak_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5, dash=[2, 8]))
            self.sweep_main_peak_line.hide()
            self.sweep_plot.addItem(self.sweep_main_peak_line)

            self.sweep_smooth_max = et.utils.SmoothMax()

            # Utilization level plot

            self.rect_plot = win.addPlot(row=1, col=0, title="Utilization level")
            self.rect_plot.setAspectLocked()
            self.rect_plot.hideAxis("left")
            self.rect_plot.hideAxis("bottom")
            self.rects = []

            pen = pg.mkPen(None)
            rect_height = self.num_rects / 2.0
            for r in np.arange(self.num_rects) + 1:
                rect = pg.QtWidgets.QGraphicsRectItem(r, rect_height, 1, rect_height)
                rect.setPen(pen)
                rect.setBrush(et.utils.pg_brush_cycler(7))
                self.rect_plot.addItem(rect)
                self.rects.append(rect)

            self.level_html_format = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:12pt;">'
                "{}</span></div>"
            )

            self.level_text_item = pg.TextItem(
                fill=pg.mkColor(0, 150, 0),
                anchor=(0.5, 0),
            )

            no_detection_html = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:12pt;">'
                "{}</span></div>".format("No level detected")
            )

            self.no_detection_text_item = pg.TextItem(
                html=no_detection_html,
                fill=pg.mkColor("r"),
                anchor=(0.5, 0),
            )

            self.rect_plot.addItem(self.level_text_item)
            self.rect_plot.addItem(self.no_detection_text_item)
            self.level_text_item.setPos(self.num_rects / 2.0 + 1, self.num_rects + 2)
            self.level_text_item.hide()
            self.no_detection_text_item.setPos(self.num_rects / 2.0 + 1, self.num_rects + 2)
            self.no_detection_text_item.show()

        # Presence

        if self.ex_app_config.activate_presence:
            assert self.cargo_presence_config is not None
            assert estimated_frame_rate is not None

            self.history_length_n = int(round(PRESENCE_RUN_TIME_S * estimated_frame_rate) + 1)
            self.intra_history = np.zeros(self.history_length_n)
            self.inter_history = np.zeros(self.history_length_n)

            # Presence history plot

            self.presence_hist_plot = win.addPlot(
                row=0,
                col=1,
                title="Presence history",
            )
            self.presence_hist_plot.setMenuEnabled(False)
            self.presence_hist_plot.setMouseEnabled(x=False, y=False)
            self.presence_hist_plot.hideButtons()
            self.presence_hist_plot.showGrid(x=True, y=True)
            self.presence_hist_plot.setLabel("bottom", "Time (s)")
            self.presence_hist_plot.setLabel("left", "Score")
            self.presence_hist_plot.setXRange(-PRESENCE_RUN_TIME_S, 0)
            self.presence_history_smooth_max = et.utils.SmoothMax(estimated_frame_rate)
            self.presence_hist_plot.setYRange(0, 10)

            self.intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            self.intra_pen = et.utils.pg_pen_cycler(1)

            self.intra_hist_curve = self.presence_hist_plot.plot(pen=self.intra_pen)
            self.intra_limit_line = pg.InfiniteLine(angle=0, pen=self.intra_dashed_pen)
            self.presence_hist_plot.addItem(self.intra_limit_line)
            self.intra_limit_line.setPos(self.cargo_presence_config.intra_detection_threshold)
            self.intra_limit_line.setPen(self.intra_dashed_pen)

            self.inter_pen = et.utils.pg_pen_cycler(0)
            self.inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

            self.inter_hist_curve = self.presence_hist_plot.plot(pen=self.inter_pen)
            self.inter_limit_line = pg.InfiniteLine(angle=0, pen=self.inter_dashed_pen)
            self.presence_hist_plot.addItem(self.inter_limit_line)
            self.inter_limit_line.setPos(self.cargo_presence_config.inter_detection_threshold)
            self.inter_limit_line.setPen(self.inter_dashed_pen)

            self.hist_xs = np.linspace(-PRESENCE_RUN_TIME_S, 0, self.history_length_n)

            # Presence detection plot

            self.presence_detection_plot = win.addPlot(row=1, col=1, title="Presence")
            self.presence_detection_plot.setAspectLocked()
            self.presence_detection_plot.hideAxis("left")
            self.presence_detection_plot.hideAxis("bottom")

            present_html_format = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:12pt;">'
                "{}</span></div>".format("Presence detected")
            )
            not_present_html = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:12pt;">'
                "{}</span></div>".format("No presence detected")
            )
            self.present_text_item = pg.TextItem(
                html=present_html_format,
                fill=pg.mkColor(0, 150, 0),
                anchor=(0.65, 0),
            )
            self.not_present_text_item = pg.TextItem(
                html=not_present_html,
                fill=pg.mkColor("r"),
                anchor=(0.6, 0),
            )

            self.presence_detection_plot.addItem(self.present_text_item)
            self.presence_detection_plot.addItem(self.not_present_text_item)
            self.present_text_item.setPos(self.num_rects / 2.0 + 1, self.num_rects + 2)
            self.not_present_text_item.setPos(self.num_rects / 2.0 + 1, self.num_rects + 2)
            self.present_text_item.hide()

            pen = pg.mkPen(None)
            rect_height = self.num_rects / 2.0
            rect_length = self.num_rects

            self.presence_rect = pg.QtWidgets.QGraphicsRectItem(
                0, rect_height, rect_length, rect_height
            )
            self.presence_rect.setPen(pen)
            self.presence_rect.setBrush(et.utils.pg_brush_cycler(7))
            self.presence_detection_plot.addItem(self.presence_rect)

    def draw_plot_job(self, *, result: ExAppResult) -> None:
        if result.mode == _Mode.DISTANCE:
            assert self.utilization_level_config is not None

            if self.ex_app_config.activate_presence:
                self.intra_history = np.zeros(self.history_length_n)
                self.inter_history = np.zeros(self.history_length_n)

            # Sweep plot

            distance = result.distance
            max_val_in_plot = 0
            if result.distance_processor_result is not None:
                for idx, processor_result in enumerate(result.distance_processor_result):
                    assert processor_result.extra_result.used_threshold is not None
                    assert processor_result.extra_result.distances_m is not None
                    assert processor_result.extra_result.abs_sweep is not None

                    abs_sweep = processor_result.extra_result.abs_sweep
                    threshold = processor_result.extra_result.used_threshold
                    distances_m = processor_result.extra_result.distances_m

                    self.sweep_curves[idx].setData(distances_m, abs_sweep)
                    self.threshold_curves[idx].setData(distances_m, threshold)

                    max_val_in_subsweep = max(max(threshold), max(abs_sweep))
                    if max_val_in_plot < max_val_in_subsweep:
                        max_val_in_plot = max_val_in_subsweep

                self.sweep_plot.setYRange(0, self.sweep_smooth_max.update(max_val_in_plot))

            if distance is not None:
                text_y_pos = self.sweep_plot.getAxis("left").range[1] * 0.95
                text_x_pos = (
                    self.sweep_plot.getAxis("bottom").range[1]
                    + self.sweep_plot.getAxis("bottom").range[0]
                ) / 2.0
                self.sweep_text_item.setPos(text_x_pos, text_y_pos)
                self.sweep_text_item.setHtml("Main peak distance: {:.3f} m".format(distance))
                self.sweep_text_item.show()

                self.sweep_main_peak_line.setPos(distance)
                self.sweep_main_peak_line.show()
            else:
                self.sweep_text_item.hide()
                self.sweep_main_peak_line.hide()

            # Utilization level plot

            # Show the percentage level plot if the plot width is greater than 400 pixels,
            # otherwise display the level as text.
            if self.plot_layout.width() < 400:
                if result.level_percent is None:
                    self.level_text_item.hide()
                    self.no_detection_text_item.show()
                elif result.level_percent > 0:
                    assert result.level_m is not None
                    assert result.level_percent is not None
                    level_text = "Utilization level: {:.1f} m, {:.0f} %".format(
                        result.level_m,
                        result.level_percent,
                    )
                    level_html = self.level_html_format.format(level_text)
                    self.level_text_item.setHtml(level_html)
                    self.level_text_item.show()
                    self.no_detection_text_item.hide()
                else:  # No detection
                    self.level_text_item.hide()
                    self.no_detection_text_item.show()

                for rect in self.rects:
                    rect.setVisible(False)
            else:
                if result.level_percent is None:  # No detection
                    for rect in self.rects:
                        rect.setBrush(et.utils.pg_brush_cycler(7))
                    self.level_text_item.hide()
                    self.no_detection_text_item.show()
                else:
                    self.bar_loc = int(
                        np.around(self.num_rects - result.level_percent / 100 * self.num_rects)
                    )
                    for rect in self.rects[self.bar_loc :]:
                        rect.setBrush(et.utils.pg_brush_cycler(0))

                    for rect in self.rects[: self.bar_loc]:
                        rect.setBrush(et.utils.pg_brush_cycler(7))

                    assert result.level_m is not None
                    assert result.level_percent is not None
                    level_text = "Utilization level: {:.2f} m, {:.0f} %".format(
                        result.level_m,
                        result.level_percent,
                    )
                    level_html = self.level_html_format.format(level_text)
                    self.level_text_item.setHtml(level_html)
                    self.level_text_item.show()
                    self.no_detection_text_item.hide()

                for rect in self.rects:
                    rect.setVisible(True)

        else:
            assert self.cargo_presence_config is not None

            # Presence history

            self.intra_history = np.roll(self.intra_history, -1)
            self.intra_history[-1] = result.intra_presence_score
            self.intra_hist_curve.setData(self.hist_xs, self.intra_history)

            self.inter_history = np.roll(self.inter_history, -1)
            self.inter_history[-1] = result.inter_presence_score
            self.inter_hist_curve.setData(self.hist_xs, self.inter_history)

            # Set y-range

            if np.isnan(self.intra_history).all():
                intra_m_hist = self.cargo_presence_config.intra_detection_threshold
            else:
                intra_m_hist = max(
                    float(np.nanmax(self.intra_history)),
                    self.cargo_presence_config.intra_detection_threshold * 1.05,
                )

            if np.isnan(self.inter_history).all():
                inter_m_hist = self.cargo_presence_config.inter_detection_threshold
            else:
                inter_m_hist = max(
                    float(np.nanmax(self.inter_history)),
                    self.cargo_presence_config.inter_detection_threshold * 1.05,
                )

            m_hist = max(intra_m_hist, inter_m_hist)
            m_hist = self.presence_history_smooth_max.update(m_hist)
            self.presence_hist_plot.setYRange(0, m_hist)

            # Presence detection plot

            if result.presence_detected:
                self.present_text_item.show()
                self.not_present_text_item.hide()
                self.presence_rect.setBrush(pg.mkColor(0, 150, 0))
            else:
                self.present_text_item.hide()
                self.not_present_text_item.show()
                self.presence_rect.setBrush(et.utils.pg_brush_cycler(7))


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor(
            title="Ex App parameters",
            factory_mapping={
                "container_size": pidgets.EnumPidgetFactory(
                    name_label_text="Container size:",
                    name_label_tooltip=get_attribute_docstring(ExAppConfig, "container_size"),
                    enum_type=ContainerSize,
                    label_mapping={
                        ContainerSize.CONTAINER_10_FT: "10 feet (3 m)",
                        ContainerSize.CONTAINER_20_FT: "20 feet (6 m)",
                        ContainerSize.CONTAINER_40_FT: "40 feet (12 m)",
                    },
                ),
                "activate_presence": pidgets.CheckboxPidgetFactory(
                    name_label_text="Activate presence detection",
                    name_label_tooltip=get_attribute_docstring(ExAppConfig, "activate_presence"),
                ),
                "activate_utilization_level": pidgets.CheckboxPidgetFactory(
                    name_label_text="Activate utilization level",
                    name_label_tooltip=get_attribute_docstring(
                        ExAppConfig, "activate_utilization_level"
                    ),
                ),
            },
            config_type=ExAppConfig,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.subsweep_status_utilization_level = SensorConfigEditor()
        self.subsweep_status_utilization_level.set_read_only(True)
        utiilization_level_status_widget = CollapsibleWidget(
            "Utilization level subsweep settings",
            self.subsweep_status_utilization_level.subsweep_group_box,
            self.scrolly_widget,
        )

        scrolly_layout.addWidget(utiilization_level_status_widget)

        self.utilization_level_config_editor = AttrsConfigEditor(
            title="Utilization level config parameters",
            factory_mapping=self._get_utilization_level_pidget_mapping(),
            config_type=UtilizationLevelConfig,
            parent=self.scrolly_widget,
        )
        self.utilization_level_config_editor.sig_update.connect(
            self._on_utilization_level_config_update
        )
        scrolly_layout.addWidget(self.utilization_level_config_editor)

        self.subsweep_status_cargo_presence = SensorConfigEditor()
        self.subsweep_status_cargo_presence.set_read_only(True)
        cargo_presence_status_widget = CollapsibleWidget(
            "Cargo presence subsweep settings",
            self.subsweep_status_cargo_presence.subsweep_group_box,
            self.scrolly_widget,
        )

        scrolly_layout.addWidget(cargo_presence_status_widget)
        self.cargo_presence_config_editor = AttrsConfigEditor(
            title="Cargo presence config parameters",
            factory_mapping=self._get_cargo_presence_pidget_mapping(),
            config_type=CargoPresenceConfig,
            parent=self.scrolly_widget,
        )
        self.cargo_presence_config_editor.sig_update.connect(self._on_cargo_presence_config_update)
        scrolly_layout.addWidget(self.cargo_presence_config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_utilization_level_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "update_rate": pidgets.FloatPidgetFactory(
                name_label_text="Update rate:",
                name_label_tooltip=get_attribute_docstring(UtilizationLevelConfig, "update_rate"),
                suffix=" Hz",
                decimals=2,
                limits=(0.1, None),
            ),
            "signal_quality": pidgets.FloatSliderPidgetFactory(
                name_label_text="Signal quality:",
                name_label_tooltip=get_attribute_docstring(
                    UtilizationLevelConfig, "signal_quality"
                ),
                decimals=1,
                limits=(-10, 35),
                show_limit_values=False,
                limit_texts=("Less power", "Higher quality"),
            ),
            "threshold_sensitivity": pidgets.FloatSliderPidgetFactory(
                name_label_text="Threshold sensitivity:",
                name_label_tooltip=get_attribute_docstring(
                    UtilizationLevelConfig, "threshold_sensitivity"
                ),
                decimals=2,
                limits=(0, 1),
                show_limit_values=False,
            ),
        }

    @classmethod
    def _get_cargo_presence_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "burst_rate": pidgets.FloatPidgetFactory(
                name_label_text="Burst rate:",
                name_label_tooltip=get_attribute_docstring(CargoPresenceConfig, "burst_rate"),
                suffix=" Hz",
                decimals=2,
                limits=(0.01, 0.2),
            ),
            "update_rate": pidgets.FloatPidgetFactory(
                name_label_text="Update rate:",
                name_label_tooltip=get_attribute_docstring(CargoPresenceConfig, "update_rate"),
                suffix=" Hz",
                decimals=1,
                limits=(1, 100),
            ),
            "signal_quality": pidgets.FloatSliderPidgetFactory(
                name_label_text="Signal quality:",
                name_label_tooltip=get_attribute_docstring(CargoPresenceConfig, "signal_quality"),
                decimals=1,
                limits=(-10.0, 60.0),
            ),
            "sweeps_per_frame": pidgets.IntPidgetFactory(
                name_label_text="Sweeps per frame:",
                name_label_tooltip=get_attribute_docstring(
                    CargoPresenceConfig, "sweeps_per_frame"
                ),
                limits=(1, 4095),
            ),
            "inter_detection_threshold": pidgets.FloatSliderPidgetFactory(
                name_label_text="Inter detection threshold:",
                name_label_tooltip=get_attribute_docstring(
                    CargoPresenceConfig, "inter_detection_threshold"
                ),
                decimals=2,
                limits=(1, 20),
                log_scale=True,
            ),
            "intra_detection_threshold": pidgets.FloatSliderPidgetFactory(
                name_label_text="Intra detection threshold:",
                name_label_tooltip=get_attribute_docstring(
                    CargoPresenceConfig, "intra_detection_threshold"
                ),
                decimals=2,
                limits=(1, 20),
                log_scale=True,
            ),
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.utilization_level_config_editor.set_data(None)
            self.cargo_presence_config_editor.set_data(None)
            self.subsweep_status_utilization_level.set_data(None)
            self.subsweep_status_cargo_presence.set_data(None)
        else:
            self.sensor_id_pidget.set_data(state.sensor_id)
            self.config_editor.set_data(state.config)

            if state.config.utilization_level_config is not None:
                self.utilization_level_config_editor.set_data(
                    state.config.utilization_level_config
                )
                self.utilization_level_config_editor.setHidden(
                    not state.config.activate_utilization_level
                )
                start_m, end_m = state.config.container_size.distance_range
                distance_session_config = detector_config_to_session_config(
                    ExApp.get_distance_config(
                        state.config.utilization_level_config, start_m, end_m
                    ),
                    [state.sensor_id],
                )
                groups = distance_session_config.groups
                sensor_config = groups[0][state.sensor_id]
                self.subsweep_status_utilization_level.set_data(sensor_config)

            if state.config.cargo_presence_config is not None:
                self.cargo_presence_config_editor.set_data(state.config.cargo_presence_config)
                self.cargo_presence_config_editor.setHidden(not state.config.activate_presence)

                start_m, end_m = state.config.container_size.presence_range
                self.subsweep_status_cargo_presence.set_data(
                    PresenceDetector._get_sensor_config(
                        ExApp.get_presence_config(
                            state.config.cargo_presence_config, start_m, end_m
                        )
                    )
                )

            results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.cargo_presence_config_editor.handle_validation_results(not_handled)

            not_handled = self.utilization_level_config_editor.handle_validation_results(
                not_handled
            )

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.config_editor,
                self.utilization_level_config_editor,
                self.cargo_presence_config_editor,
                self.sensor_id_pidget,
            ],
        )

        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=self.config_editor.is_ready
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: ExAppConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_utilization_level_config_update(self, config: UtilizationLevelConfig) -> None:
        BackendPlugin.update_utilization_level_config.rpc(self.app_model.put_task, config=config)

    def _on_cargo_presence_config_update(self, config: CargoPresenceConfig) -> None:
        BackendPlugin.update_cargo_presence_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


CARGO_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="cargo",
    title="Cargo",
    docs_link="https://docs.acconeer.com/en/latest/example_apps/a121/cargo.html",
    description="Detects utilization level and presence in container.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(
            name="10 feet (3 m)",
            description="Container 10 feet (3 m)",
            preset_id=PluginPresetId.CONTAINER_10_FT,
        ),
        PluginPresetBase(
            name="20 feet (6 m)",
            description="Container 20 feet (6 m)",
            preset_id=PluginPresetId.CONTAINER_20_FT,
        ),
        PluginPresetBase(
            name="40 feet (12 m)",
            description="Container 40 feet (12 m)",
            preset_id=PluginPresetId.CONTAINER_40_FT,
        ),
        PluginPresetBase(
            name="No lens",
            description="No lens configuration",
            preset_id=PluginPresetId.NO_LENS,
        ),
    ],
    default_preset_id=PluginPresetId.NO_LENS,
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import copy
import logging
from typing import Any, List, Optional

from PySide6.QtCore import QSignalBlocker, Qt, Signal
from PySide6.QtWidgets import (
    QGridLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool import a121
from acconeer.exptool._core.entities.validation_result import Criticality
from acconeer.exptool.a121._core import utils
from acconeer.exptool.app.new.ui.components import pidgets
from acconeer.exptool.app.new.ui.components.data_editor import DataEditor
from acconeer.exptool.app.new.ui.components.group_box import GroupBox
from acconeer.exptool.app.new.ui.components.json_save_load_buttons import (
    create_json_save_load_buttons,
)
from acconeer.exptool.app.new.ui.icons import PLUS, REMOVE

from .sensor_config_editor import SensorConfigEditor


log = logging.getLogger(__name__)


def session_or_sensor_config_json_to_session_config(json_str: str) -> a121.SessionConfig:
    try:
        return a121.SessionConfig.from_json(json_str)
    except KeyError:
        return a121.SessionConfig(a121.SensorConfig.from_json(json_str))


class SensorIdCombobox(DataEditor[Optional[int]]):
    sig_update = Signal(object)

    def __init__(self) -> None:
        super().__init__()

        self._selectable_sensors: list[int] = []
        self._combobox = pidgets.PidgetComboBox(self)
        self._combobox.currentIndexChanged.connect(self._emit_data_at_index)
        self.set_data(None)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._combobox)

        self.setLayout(layout)

    def setEnabled(self, enabled: bool) -> None:
        """Resets the DataEditor.setEnabled behavior to the default QWidget"""
        QWidget.setEnabled(self, enabled)

    @staticmethod
    def _get_combobox_model(
        data: Optional[int], selectable: list[int]
    ) -> list[tuple[str, Optional[int]]]:
        base: list[tuple[str, Optional[int]]] = [(str(sid), sid) for sid in selectable]
        if data is None:
            return [("-", None)] + base

        if data in selectable:
            return base
        else:
            return [(f"{data} (unavailable)", data)] + base

    def _update_combobox_model(self, data: Optional[int], selectable: list[int]) -> None:
        with QSignalBlocker(self):
            self._combobox.clear()

            for label, sid in self._get_combobox_model(data, selectable):
                self._combobox.addItem(label, sid)

            self._combobox.setCurrentIndex(self._combobox.findData(data))

    def _emit_data_at_index(self, index: int) -> None:
        self.sig_update.emit(self._combobox.itemData(index))

    def set_selectable_sensors(self, sensor_list: list[int]) -> None:
        self._selectable_sensors = sensor_list
        self._update_combobox_model(self.get_data(), sensor_list)
        self.setEnabled(sensor_list != [])

    def set_data(self, sensor_id: Optional[int]) -> None:
        with QSignalBlocker(self):
            self._combobox.setCurrentIndex(self._combobox.findData(sensor_id))
            self._update_combobox_model(sensor_id, self._selectable_sensors)

    def get_data(self) -> Optional[int]:
        try:
            return int(self._combobox.currentData())
        except TypeError:
            return None


class _SensorIdsEditor(DataEditor[List[int]]):
    _LABEL_COLUMN = 0
    _SPACING_COLUMN = 1
    _REMOVE_COLUMN = 2
    _COMBOBOX_COLUMN = 3

    sig_update = Signal(object)

    def __init__(self, multiple_sensors: bool) -> None:
        """
        Composes multiple comboboxes into a vertical layout where
        pidgets can be added or removed (except a single static combobox).

        :param multiple_sensors:
            True enables adding/removing pidgets and False gives this
            widget the appearance of a single SensorIdPidget
        """
        super().__init__()

        self._multiple_sensors = multiple_sensors
        self._selectable_sensors: list[int] = []
        self._rows: list[tuple[SensorIdCombobox, Optional[QPushButton]]] = []

        self._grid_layout = QGridLayout()
        self._grid_layout.setContentsMargins(0, 0, 0, 0)
        self._grid_layout.setColumnStretch(self._LABEL_COLUMN, 0)
        self._grid_layout.setColumnStretch(self._REMOVE_COLUMN, 0)
        self._grid_layout.setColumnStretch(self._SPACING_COLUMN, 1)
        self._grid_layout.setColumnStretch(self._COMBOBOX_COLUMN, 0)

        self._add_button = QPushButton()
        self._add_button.setIcon(PLUS())
        self._add_button.setFlat(True)
        self._add_button.setFixedWidth(pidgets.WIDGET_WIDTH)
        self._add_button.setFixedWidth(125)
        self._add_button.setVisible(multiple_sensors)
        self._add_button.clicked.connect(self._create_and_add_row)
        self._add_button.clicked.connect(self._update_visuals)
        self._add_button.clicked.connect(self._emit_current_data)

        self._note_widget = QLabel()
        self._note_widget.setWordWrap(True)
        self._note_widget.setContentsMargins(5, 5, 5, 5)
        self._note_widget.hide()

        # Static row
        static_combobox = SensorIdCombobox()
        static_combobox.sig_update.connect(self._update_visuals)
        static_combobox.sig_update.connect(self._emit_current_data)

        self._rows.append((static_combobox, None))

        self._grid_layout.addWidget(QLabel("Sensors:"), 0, self._LABEL_COLUMN)
        self._grid_layout.addWidget(static_combobox, 0, self._COMBOBOX_COLUMN)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(self._grid_layout)
        layout.addWidget(self._add_button, stretch=0, alignment=Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self._note_widget)
        self.setLayout(layout)

    @property
    def _free_sensors(self) -> list[int]:
        """Returns a list of all the sensors that are connected but not selected"""
        return [sid for sid in self._selectable_sensors if sid not in self.get_data()]

    def _emit_current_data(self) -> None:
        self.sig_update.emit(self.get_data())

    def _get_raw_data(self) -> list[Optional[int]]:
        """Retrieves the data of the child comboboxes"""
        return [cb.get_data() for cb, _ in self._rows]

    def _update_visuals(self) -> None:
        self._add_button.setDisabled(self._free_sensors == [] or None in self._get_raw_data())

        self._add_button.setToolTip(
            "No connected sensors left"
            if self._free_sensors == []
            else (
                "Make a selection in the dashed drop down before adding a new row"
                if None in self._get_raw_data()
                else ""
            )
        )

        for combobox, _ in self._rows:
            if self._selectable_sensors == []:
                combobox.set_selectable_sensors([])
            else:
                current_data = combobox.get_data()
                if current_data is None:
                    combobox.set_selectable_sensors(sorted(self._free_sensors))
                else:
                    combobox.set_selectable_sensors(sorted(self._free_sensors + [current_data]))

    def _create_and_add_row(
        self,
    ) -> None:
        row_to_append = self._grid_layout.rowCount()

        combobox = SensorIdCombobox()
        combobox.set_selectable_sensors(self._free_sensors)
        combobox.sig_update.connect(self._update_visuals)
        combobox.sig_update.connect(self._emit_current_data)

        remove_button = QPushButton()
        remove_button.setIcon(REMOVE())
        remove_button.setToolTip("Removes this sensor")
        remove_button.setFlat(True)
        remove_button.clicked.connect(lambda: self._rows.remove((combobox, remove_button)))
        remove_button.clicked.connect(self._update_visuals)
        remove_button.clicked.connect(self._emit_current_data)
        remove_button.clicked.connect(lambda: combobox.deleteLater())
        remove_button.clicked.connect(lambda: remove_button.deleteLater())

        self._grid_layout.addWidget(combobox, row_to_append, self._COMBOBOX_COLUMN)
        self._grid_layout.addWidget(remove_button, row_to_append, self._REMOVE_COLUMN)

        self._rows.append((combobox, remove_button))

    def set_data(self, data: list[int]) -> None:
        row_balance = len(self._rows) - len(data)

        if row_balance < 0:
            num_missing_rows = -row_balance

            for _ in range(num_missing_rows):
                self._create_and_add_row()

        if row_balance > 0:
            num_redundant_rows = row_balance

            for (_, remove_button), _ in zip(reversed(self._rows), range(num_redundant_rows)):
                if remove_button is not None:
                    remove_button.click()

        for sid, (combobox, _) in zip(data, self._rows):
            combobox.set_data(sid)

    def get_data(self) -> list[int]:
        """The list of currently selected sensor ids"""
        return [sid for sid in self._get_raw_data() if sid is not None]

    def set_selectable_sensors(self, sensors: list[int]) -> None:
        self._selectable_sensors = sensors
        self._update_visuals()

    def set_note_text(
        self, message: Optional[str], criticality: Optional[Criticality] = None
    ) -> None:
        if not message:
            self._note_widget.hide()
            return

        COLOR_MAP = {
            Criticality.ERROR: "#E6635A",
            Criticality.WARNING: "#FCC842",
            None: "white",
        }

        self._note_widget.show()
        self._note_widget.setText(message)
        self._note_widget.setStyleSheet(
            f"background-color: {COLOR_MAP[criticality]}; color: white; font: bold italic;"
        )

    @property
    def is_ready(self) -> bool:
        return self._note_widget.isHidden()


class SessionConfigEditor(DataEditor[Optional[a121.SessionConfig]]):
    _session_config: Optional[a121.SessionConfig]
    _server_info: Optional[a121.ServerInfo]
    _sensor_id_pidget: pidgets.SensorIdPidget

    _update_rate_erroneous: bool

    sig_update = Signal(object)

    SPACING = 15

    def __init__(
        self,
        supports_multiple_subsweeps: bool = False,
        supports_multiple_sensors: bool = False,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)

        self._server_info = None

        self._session_config = None
        self._update_rate_erroneous = False

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.session_group_box = GroupBox.vertical(
            "Session parameters",
            create_json_save_load_buttons(
                self,
                encoder=a121.SessionConfig.to_json,
                decoder=session_or_sensor_config_json_to_session_config,
            ),
            parent=self,
        )
        self.session_group_box.layout().setSpacing(self.SPACING)

        self._sensor_ids_editor = _SensorIdsEditor(supports_multiple_sensors)

        self._sensor_ids_editor.sig_update.connect(self._update_sensor_ids)

        self.session_group_box.layout().addWidget(self._sensor_ids_editor)

        layout.addWidget(self.session_group_box)

        self._update_rate_pidget = pidgets.OptionalFloatPidgetFactory(
            name_label_text="Update rate:",
            name_label_tooltip=(
                "Set an update rate limit on the server.\n"
                "If 'Limit' is unchecked, the server will run as fast as possible."
            ),
            limits=(0.1, 1e4),
            decimals=1,
            init_set_value=10.0,
            placeholder_text="- Hz",
            suffix="Hz",
            checkbox_label_text="Limit",
        ).create(self)
        self._update_rate_pidget.sig_update.connect(self._update_update_rate)
        self.session_group_box.layout().addWidget(self._update_rate_pidget)

        self._sensor_config_editor = SensorConfigEditor(supports_multiple_subsweeps, parent=self)
        self._sensor_config_editor.sig_update.connect(self._update_sole_sensor_config)
        layout.addWidget(self._sensor_config_editor)

        self.setLayout(layout)

    def set_selectable_sensors(self, sensor_list: list[int]) -> None:
        self._sensor_ids_editor.set_selectable_sensors(sensor_list)

    def set_read_only(self, read_only: bool) -> None:
        self._sensor_ids_editor.setEnabled(not read_only)
        self._update_rate_pidget.setEnabled(not read_only)
        self._sensor_config_editor.set_read_only(read_only)

    def set_data(self, session_config: Optional[a121.SessionConfig]) -> None:
        if self._session_config == session_config:
            return

        self._session_config = session_config
        self.setEnabled(session_config is not None)

        if self._session_config is None:
            log.debug("could not update ui as SessionConfig is None")
            return

        self._sensor_ids_editor.set_data(self._get_sensor_ids(self._session_config))
        self._update_rate_pidget.set_data(self._session_config.update_rate)
        self._sensor_config_editor.set_data(self._get_unique_sensor_config(self._session_config))

    @staticmethod
    def _get_sensor_ids(config: a121.SessionConfig) -> list[int]:
        return [sensor_id for _, sensor_id, _ in utils.iterate_extended_structure(config.groups)]

    def get_data(self) -> Optional[a121.SessionConfig]:
        return copy.deepcopy(self._session_config)

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        self._update_rate_pidget.set_note_text(None)
        self._sensor_ids_editor.set_note_text(None)

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        unhandled_results = self._sensor_config_editor.handle_validation_results(unhandled_results)

        return unhandled_results

    @property
    def is_ready(self) -> bool:
        return not self._update_rate_erroneous and self._sensor_config_editor.is_ready

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if self._session_config is None:
            msg = "SessionConfigEditor's config is None while ValidationResults are being handled."
            raise RuntimeError(msg)

        result_handled = False

        if result.aspect == "update_rate":
            self._update_rate_pidget.set_note_text(result.message, result.criticality)
            result_handled = True
        elif result.aspect == "sensor_id":
            self._sensor_ids_editor.set_note_text(result.message, result.criticality)
            result_handled = True

        return result_handled

    def _update_update_rate(self, value: Any) -> None:
        if self._session_config is None:
            msg = "SessionConfig is None"
            raise TypeError(msg)

        config = copy.deepcopy(self._session_config)

        try:
            config.update_rate = value
        except Exception as e:
            self._update_rate_erroneous = True
            self._update_rate_pidget.set_note_text(e.args[0], Criticality.ERROR)

            # this emit needs to be done to signal that "is_ready" has changed.
            self.sig_update.emit(self.get_data())
        else:
            self._update_rate_erroneous = False

            self.set_data(config)
            self.sig_update.emit(config)

    def _update_sensor_ids(self, sensor_ids: list[int]) -> None:
        if self._session_config is None:
            msg = "SessionConfig is None"
            raise TypeError(msg)

        config = copy.deepcopy(self._session_config)

        sensor_config = self._get_unique_sensor_config(config)

        new_config = a121.SessionConfig(
            self._assemble_groups(sensor_ids, sensor_config),
            update_rate=config.update_rate,
        )

        self.set_data(new_config)
        self.sig_update.emit(new_config)

    @staticmethod
    def _get_unique_sensor_config(config: a121.SessionConfig) -> a121.SensorConfig:
        (first_config, *other_configs) = utils.iterate_extended_structure_values(config.groups)

        if any(sc != first_config for sc in other_configs):
            msg = "SessionConfig contains more than 1 unique SensorConfig"
            raise ValueError(msg)

        return first_config

    @staticmethod
    def _assemble_groups(
        sensor_ids: list[int], sensor_config: a121.SensorConfig
    ) -> list[dict[int, a121.SensorConfig]]:
        return [{sensor_id: sensor_config for sensor_id in sensor_ids}]

    def _update_sole_sensor_config(self, sensor_config: a121.SensorConfig) -> None:
        if self._session_config is None:
            raise RuntimeError

        config = a121.SessionConfig(
            self._assemble_groups(self._get_sensor_ids(self._session_config), sensor_config),
            update_rate=self._session_config.update_rate,
        )
        self.set_data(config)
        self.sig_update.emit(config)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import enum
import json
import traceback
import typing as t
from pathlib import Path

import typing_extensions as te

from PySide6.QtCore import QSize, QTimer
from PySide6.QtWidgets import QHBoxLayout, QPushButton, QWidget

from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.app.new.ui.components.data_editor import DataEditor
from acconeer.exptool.app.new.ui.icons import CHECKMARK, FOLDER_OPEN, SAVE
from acconeer.exptool.app.new.ui.misc import ExceptionWidget

from .dialogs import (
    LoadDialogWithJsonEditor,
    PresentationType,
    PresenterFunc,
    SaveDialogWithPreview,
    set_config_presenter,
)


_ICON_SIZE = QSize(20, 20)
_BUTTON_SIZE = QSize(35, 25)
_SAVE_FILE_DIALOG_FILTER = "JSON (*.json);;CSV (*.csv)"
_LOAD_FILE_DIALOG_FILTER = "JSON (*.json)"


@te.runtime_checkable
class JsonPresentable(te.Protocol):
    def to_json(self) -> str: ...

    @classmethod
    def from_json(cls, json_str: str) -> te.Self: ...


_JsonPresentableT = t.TypeVar("_JsonPresentableT", bound=JsonPresentable)


def _none_coalescing_chain(*funcs: PresenterFunc) -> PresenterFunc:
    """
    Composes a sequence of PresenterFuncs into a chain.

    The chain will call the functions in order,
    terminating at (and returning the result of)
    the first function returning a non-None value.
    """

    def _chain(instance: t.Any, t: PresentationType) -> t.Optional[str]:
        for f in funcs:
            presentation = f(instance, t)
            if presentation is not None:
                return presentation
        return None

    return _chain


def _json_presentation(instance: t.Any, t: PresentationType) -> t.Optional[str]:
    if t is PresentationType.JSON and isinstance(instance, JsonPresentable):
        return json.dumps(
            json.loads(instance.to_json()),
            indent=2,
        )

    return None


def _show_transient_checkmark(button: QPushButton, duration_ms: int = 2000) -> None:
    """Temporarily changes the icon of button to a checkmark"""
    original_icon = button.icon()
    button.setIcon(CHECKMARK())
    QTimer.singleShot(duration_ms, lambda: button.setIcon(original_icon))


def _handle_exception_with_popup(
    f: t.Callable[..., t.Any], header: str, parent: QWidget
) -> t.Callable[..., t.Any]:
    """Decorates the function "f" by displaying any raised exception in an ExceptionWidget"""

    def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
        try:
            return f(*args, **kwargs)
        except Exception as e:
            ExceptionWidget(
                parent=parent,
                exc=HandledException(
                    header + "\n" + f"The following exception was raised: {e!r}\n"
                ),
                traceback_str=traceback.format_exc(),
            ).exec()

    return wrapper


class _JsonLoadButton(QPushButton):
    def __init__(
        self,
        setter: t.Callable[[_JsonPresentableT], t.Any],
        decoder: t.Callable[[str], _JsonPresentableT],
        parent: t.Optional[QWidget] = None,
    ) -> None:
        super().__init__(FOLDER_OPEN(), "", parent=parent)

        self._setter = setter
        self._decoder = decoder

        self.setFixedSize(_BUTTON_SIZE)
        self.setIconSize(_ICON_SIZE)
        self.setToolTip("Load a config from file/clipboard")
        self.clicked.connect(
            _handle_exception_with_popup(
                self._load_and_set_selected_file,
                header="Could not load config from file/clipboard",
                parent=self,
            )
        )

    def _load_and_set_selected_file(self) -> None:
        contents = LoadDialogWithJsonEditor.get_load_contents(
            caption="Load from file",
            filter=_LOAD_FILE_DIALOG_FILTER,
            parent=self,
        )

        if contents:
            self._setter(self._decoder(contents))
            _show_transient_checkmark(self)


class _JsonSaveButton(QPushButton):
    def __init__(
        self,
        getter: t.Callable[[], t.Optional[_JsonPresentableT]],
        encoder: t.Callable[[_JsonPresentableT], str],
        extra_presenter: PresenterFunc,
    ) -> None:
        super().__init__(SAVE(), "")
        self._getter = getter
        self._encoder = encoder
        self._extra_presenter = extra_presenter

        self.setFixedSize(_BUTTON_SIZE)
        self.setIconSize(_ICON_SIZE)
        self.setToolTip("Save the current config to file")
        self.clicked.connect(
            _handle_exception_with_popup(
                self._save_to_selected_file,
                header="Could not save config to file.",
                parent=self,
            )
        )

    def _save_to_selected_file(self) -> None:
        model = self._getter()

        if model is None:
            raise RuntimeError

        filename, file_filter = SaveDialogWithPreview.get_save_file_name(
            caption="Save to file",
            model=model,
            presenter=_none_coalescing_chain(
                _json_presentation,
                set_config_presenter,
                self._extra_presenter,
            ),
            filter=_SAVE_FILE_DIALOG_FILTER,
            parent=self,
        )

        if filename and file_filter:
            if "json" in file_filter:
                Path(filename).with_suffix(".json").write_text(self._encoder(model))
            elif "csv" in file_filter:
                Path(filename).with_suffix(".csv").write_text(
                    _dict_to_csv(_flatten_dict(json.loads(self._encoder(model))))
                )
            else:
                msg = f"Unknown file type {file_filter}"
                raise TypeError(msg)
            _show_transient_checkmark(self)


def _dict_to_csv(data: t.Dict[t.Any, t.Any]) -> str:
    output_str = ""

    for key, value in data.items():
        output_str += f"{key},{value}\n"

    return output_str


def _flatten_dict(
    data: t.Dict[str, t.Any], root: t.Optional[t.Dict[str, t.Any]] = None, parent_key: str = ""
) -> t.Dict[str, t.Any]:
    if root is None:
        root = dict()

    for key, value in data.items():
        src_key = parent_key + "_" + key if parent_key else key
        if isinstance(value, dict):
            _flatten_dict(value, root, src_key)
        elif isinstance(value, list):
            for i, v in enumerate(value):
                if src_key == "groups":
                    # In SessionConfig groups are a list of dicts with a single key,
                    # which corresponds to the number of the group.
                    # So no need to enumerate these keys.
                    sub_key = src_key
                else:
                    sub_key = src_key + "_" + str(i)

                if isinstance(v, dict):
                    _flatten_dict(v, root, sub_key)
                else:
                    root[sub_key] = v if v is not None else "None"
        else:
            root[src_key] = value if value is not None else "None"

    return root


def create_json_save_load_buttons(
    editor: DataEditor[t.Optional[_JsonPresentableT]],
    encoder: t.Optional[t.Callable[[_JsonPresentableT], str]] = None,
    decoder: t.Optional[t.Callable[[str], _JsonPresentableT]] = None,
    extra_presenter: PresenterFunc = lambda i, t: None,
) -> QWidget:
    """
    Creates buttons for saving/loading configs to/from json.

    :param editor: The editor to bind the save/load buttons to
    :param encoder:
        A function that formats the model as a json string.
        If omitted or None, the Save button will not be created.
    :param decoder:
        A function that parses a json string produced by encoder
        If omitted or None, the Load button will not be created.
    :param extra_presenter:
        A PresenterFunc hooks into save preview presentation creation.
        This function should always return None if its arguments aren't handled.
    """
    wrapper = QWidget()
    wrapper_layout = QHBoxLayout()
    wrapper.setLayout(wrapper_layout)
    wrapper_layout.setContentsMargins(5, 5, 5, 8)

    if encoder is not None:
        wrapper_layout.addWidget(
            _JsonSaveButton(
                editor.get_data,
                encoder,
                extra_presenter=extra_presenter,
            )
        )

    def setter(data: _JsonPresentableT) -> t.Any:
        editor.set_data(data)
        editor.sig_update.emit(data)

    if decoder is not None:
        wrapper_layout.addWidget(_JsonLoadButton(setter, decoder))

    return wrapper


class JsonButtonOperations(enum.Flag):
    SAVE = enum.auto()
    LOAD = enum.auto()


def create_json_save_load_buttons_from_type(
    editor: DataEditor[t.Optional[_JsonPresentableT]],
    config_type: t.Type[_JsonPresentableT],
    operations: JsonButtonOperations = JsonButtonOperations.SAVE | JsonButtonOperations.LOAD,
    extra_presenter: PresenterFunc = lambda i, t: None,
) -> t.Optional[QWidget]:
    """
    Utility variation of "create_json_save_load_buttons".

    :param config_type: The type (class) of the config
    :param operations:
        Defines what operations the button(s) should perform.
        This directly maps to which buttons are created.
        If no operations are specified, None will be returned.

    See create_json_save_load_buttons for the other parameters.
    """
    if operations == JsonButtonOperations(0):
        return None
    else:
        return create_json_save_load_buttons(
            editor,
            encoder=config_type.to_json if (operations & JsonButtonOperations.SAVE) else None,
            decoder=config_type.from_json if (operations & JsonButtonOperations.LOAD) else None,
            extra_presenter=extra_presenter,
        )



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

from acconeer.exptool.setup.base import PlatformInstall, SetupGroup, ShellCommandStep, utils

from .linux import Linux


@PlatformInstall.register
class Ubuntu_22_04(Linux):
    def __init__(self) -> None:
        super().__init__()
        self.add_steps(
            utils.WithDescription(
                '> Download dependencies via "apt" that are required '
                + "to run the Exploration tool app.",
                SetupGroup(
                    ShellCommandStep("sudo apt update".split()),
                    ShellCommandStep(
                        "sudo apt install -y libxcb-xinerama0-dev libusb-1.0-0 libxcb-cursor0".split()
                    ),
                ),
            ),
            utils.WithDescription(
                '> Use "apt" to remove modemmanager, this is needed to not interfere with bootloader for some EVKs.',
                SetupGroup(
                    ShellCommandStep("sudo apt purge -y modemmanager".split()),
                ),
            ),
            utils.WithDescription(
                "> Trigger udev to update the permission rules.",
                SetupGroup(
                    ShellCommandStep("sudo udevadm control --reload-rules".split()),
                    ShellCommandStep("sudo udevadm trigger".split()),
                ),
            ),
        )

    @classmethod
    def get_key(cls) -> str:
        return "Ubuntu_22.04"



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt
from numpy import cos, pi, sqrt, square
from scipy.special import binom

from acconeer.exptool import a121
from acconeer.exptool import type_migration as tm
from acconeer.exptool._core.class_creation.attrs import attrs_ndarray_isclose
from acconeer.exptool.a121.algo import AlgoProcessorConfigBase, ProcessorBase
from acconeer.exptool.a121.algo._utils import get_distances_m


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    intra_enable: bool = attrs.field(default=True)
    inter_enable: bool = attrs.field(default=True)
    intra_detection_threshold: float = attrs.field(default=1.3)
    inter_detection_threshold: float = attrs.field(default=1)
    inter_frame_presence_timeout: Optional[int] = attrs.field(default=None)
    inter_frame_fast_cutoff: float = attrs.field(default=20.0)
    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    inter_output_time_const: float = attrs.field(default=5)
    intra_frame_time_const: float = attrs.field(default=0.15)
    intra_output_time_const: float = attrs.field(default=0.5)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.frame_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "frame_rate",
                    "Must be set",
                )
            )

        if config.sensor_config.sweeps_per_frame <= Processor.NOISE_ESTIMATION_DIFF_ORDER:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweeps_per_frame",
                    f"Must be greater than {Processor.NOISE_ESTIMATION_DIFF_ORDER}",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext:
    estimated_frame_rate: Optional[float] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    frame: npt.NDArray[np.complex128] = attrs.field(eq=attrs_ndarray_isclose)
    abs_mean_sweep: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    fast_lp_mean_sweep: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    slow_lp_mean_sweep: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    lp_noise: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    presence_distance_index: int = attrs.field()


@attrs.frozen(kw_only=True)
class ProcessorResult:
    intra_presence_score: float = attrs.field()
    intra: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    inter_presence_score: float = attrs.field()
    inter: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    presence_distance: float = attrs.field()
    presence_detected: bool = attrs.field()
    extra_result: ProcessorExtraResult = attrs.field()


class Processor(ProcessorBase[ProcessorResult]):
    # lp(f): low pass (filtered)
    # cut: cutoff frequency [Hz]
    # tc: time constant [s]
    # sf: smoothing factor [dimensionless]

    MAX_AMPLITUDE_WEIGHT = 5
    NOISE_ESTIMATION_DIFF_ORDER = 3

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_indexes: Optional[list[int]] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        # Subsweep indexes contains a list of subsweep indexes for which to run the presence detector.
        # If None is supplied, use all possible.
        if subsweep_indexes is None:
            subsweep_indexes = list(range(sensor_config.num_subsweeps))
        self.subsweep_indexes = subsweep_indexes

        if context is None:
            context = ProcessorContext()

        self.sensor_config = sensor_config
        self.metadata = metadata
        self.processor_config = processor_config

        self.processor_config.validate(self.sensor_config)

        # Should never happen, checked in validate
        assert self.sensor_config.frame_rate is not None

        self.sweeps_per_frame = self.sensor_config.sweeps_per_frame
        self.distances = get_distances_m(self.sensor_config, metadata)
        self.num_distances = self.distances.size
        if context.estimated_frame_rate is not None:
            self.f = context.estimated_frame_rate
        else:
            self.f = self.sensor_config.frame_rate

        # Fixed parameters
        self.noise_est_diff_order = self.NOISE_ESTIMATION_DIFF_ORDER
        noise_tc = 10.0

        self.noise_sf = self._tc_to_sf(noise_tc, self.f)

        nd = self.noise_est_diff_order
        self.noise_norm_factor = np.sqrt(np.sum(np.square(binom(nd, np.arange(nd + 1)))))

        self.fast_lp_mean_sweep = np.zeros(self.num_distances)
        self.slow_lp_mean_sweep = np.zeros(self.num_distances)
        self.lp_inter_dev = np.zeros(self.num_distances)
        self.lp_intra_dev = np.zeros(self.num_distances)
        self.lp_noise = np.zeros(self.num_distances)

        self.intra_presence_score = 0
        self.inter_presence_score = 0
        self.presence_distance_index = 0
        self.presence_distance = 0

        self.update_index = 0

        self.intra_enable = processor_config.intra_enable
        self.intra_threshold = processor_config.intra_detection_threshold

        self.inter_enable = processor_config.inter_enable
        self.inter_threshold = processor_config.inter_detection_threshold

        self.fast_sf = self._cutoff_to_sf(processor_config.inter_frame_fast_cutoff, self.f)
        self.slow_sf = self._cutoff_to_sf(processor_config.inter_frame_slow_cutoff, self.f)
        self.inter_dev_sf = self._tc_to_sf(
            processor_config.inter_frame_deviation_time_const, self.f
        )
        self.intra_sf = self._tc_to_sf(processor_config.intra_frame_time_const, self.f)
        self.intra_output_sf = self._tc_to_sf(processor_config.intra_output_time_const, self.f)
        self.inter_output_sf = self._tc_to_sf(processor_config.inter_output_time_const, self.f)

        self.previous_presence_score = 0
        self.negative_count = 0
        self.inter_frame_presence_timeout = self.processor_config.inter_frame_presence_timeout

    @staticmethod
    def _cutoff_to_sf(fc: float, fs: float) -> float:
        """Cutoff frequency to smoothing factor conversion"""

        if fc > 0.5 * fs:
            return 0.0

        cos_w = cos(2.0 * pi * (fc / fs))
        return 2.0 - cos_w - sqrt(square(cos_w) - 4.0 * cos_w + 3.0)  # type: ignore[no-any-return]

    @staticmethod
    def _tc_to_sf(tc: float, fs: float) -> float:
        """Time constant to smoothing factor conversion"""

        if tc <= 0.0:
            return 0.0

        return np.exp(-1.0 / (tc * fs))  # type: ignore[no-any-return]

    @staticmethod
    def _dynamic_sf(static_sf: float, update_index: int) -> float:
        return min(static_sf, 1.0 - 1.0 / (1.0 + update_index))

    @staticmethod
    def _abs_dev(
        a: npt.NDArray[np.complex128],
        axis: Optional[int] = None,
        ddof: int = 0,
        subtract_mean: bool = True,
    ) -> npt.NDArray[np.float64]:
        if subtract_mean:
            a = a - a.mean(axis=axis, keepdims=True)

        if axis is None:
            n = a.size
        else:
            n = a.shape[axis]

        if ddof < 0:
            msg = "ddof must be greater than or equal to 0"
            raise ValueError(msg)

        if n <= ddof:
            msg = "n must be greater than ddof"
            raise ValueError(msg)

        return np.mean(np.abs(a), axis=axis) * sqrt(n / (n - ddof))  # type: ignore[no-any-return]

    def _inter_presence_score_scaling(self) -> None:
        """
        Scaling of self.inter_presence_score for faster decline when loosing detection.
        Start exponential scaling after self.inter_frame_presence_timeout seconds.
        """

        if self.inter_frame_presence_timeout is None:
            msg = "inter_frame_presence_timeout must be set"
            raise ValueError(msg)

        scaling_factor = np.exp(
            np.maximum(self.negative_count - self.inter_frame_presence_timeout * self.f, 0)
            / (self.inter_frame_presence_timeout * self.f)
        )
        self.inter_presence_score /= scaling_factor

    def process(self, result: a121.Result) -> ProcessorResult:
        range_subframes = [result.subframes[i] for i in self.subsweep_indexes]
        frame = np.concatenate(range_subframes, axis=1)

        # Noise estimation

        nd = self.noise_est_diff_order

        noise_diff = np.diff(frame, n=nd, axis=0)
        noise = self._abs_dev(noise_diff, axis=0, subtract_mean=False)
        noise /= self.noise_norm_factor
        sf = self._dynamic_sf(self.noise_sf, self.update_index)
        self.lp_noise = sf * self.lp_noise + (1.0 - sf) * noise

        # Intra-frame part

        sweep_dev = self._abs_dev(frame, axis=0, ddof=1)

        sf = self._dynamic_sf(self.intra_sf, self.update_index)
        self.lp_intra_dev = sf * self.lp_intra_dev + (1.0 - sf) * sweep_dev

        intra = np.divide(
            self.lp_intra_dev,
            self.lp_noise,
            out=np.zeros(self.num_distances),
            where=(self.lp_noise > 1.0),
        )

        intra_presence_distance_index = int(np.argmax(intra))
        intra_presence_distance = self.distances[intra_presence_distance_index]

        self.intra_presence_score = (
            self.intra_output_sf * self.intra_presence_score
            + (1.0 - self.intra_output_sf) * intra[intra_presence_distance_index]
        )

        # Inter-frame part

        mean_sweep = frame.mean(axis=0)
        abs_mean_sweep = np.abs(mean_sweep)

        sf = self._dynamic_sf(self.fast_sf, self.update_index)
        self.fast_lp_mean_sweep = sf * self.fast_lp_mean_sweep + (1.0 - sf) * abs_mean_sweep

        sf = self._dynamic_sf(self.slow_sf, self.update_index)
        self.slow_lp_mean_sweep = sf * self.slow_lp_mean_sweep + (1.0 - sf) * abs_mean_sweep

        inter_dev = np.abs(self.fast_lp_mean_sweep - self.slow_lp_mean_sweep)
        sf = self._dynamic_sf(self.inter_dev_sf, self.update_index)
        self.lp_inter_dev = sf * self.lp_inter_dev + (1.0 - sf) * inter_dev

        inter = np.divide(
            self.lp_inter_dev,
            self.lp_noise,
            out=np.zeros_like(self.lp_inter_dev),
            where=(self.lp_noise > 1.0),
        )

        inter *= np.sqrt(self.sweeps_per_frame)

        inter_presence_distance_index = int(np.argmax(inter))
        inter_presence_distance = self.distances[inter_presence_distance_index]

        sf = self._dynamic_sf(self.inter_output_sf, self.update_index)
        self.inter_presence_score = (
            sf * self.inter_presence_score + (1.0 - sf) * inter[inter_presence_distance_index]
        )

        # Inter-frame presence timeout

        if self.inter_frame_presence_timeout:
            delta = self.inter_presence_score - self.previous_presence_score

            if delta < 0:
                self.negative_count += 1
            else:
                self.negative_count = 0

            self._inter_presence_score_scaling()

            self.previous_presence_score = self.inter_presence_score

        # Presence distance - intra presence distance is prioritized due to faster reaction time

        if self.intra_presence_score > self.intra_threshold and self.intra_enable:
            presence_detected = True
            self.presence_distance_index = intra_presence_distance_index
            self.presence_distance = intra_presence_distance
        elif self.inter_presence_score > self.inter_threshold and self.inter_enable:
            presence_detected = True
            self.presence_distance_index = inter_presence_distance_index
            self.presence_distance = inter_presence_distance
        else:
            presence_detected = False
            self.presence_distance = 0

        self.update_index += 1

        extra_result = ProcessorExtraResult(
            frame=frame,
            abs_mean_sweep=abs_mean_sweep,
            fast_lp_mean_sweep=self.fast_lp_mean_sweep,
            slow_lp_mean_sweep=self.slow_lp_mean_sweep,
            lp_noise=self.lp_noise,
            presence_distance_index=self.presence_distance_index,
        )

        return ProcessorResult(
            intra_presence_score=self.intra_presence_score,
            intra=intra,
            inter_presence_score=self.inter_presence_score,
            inter=inter,
            presence_detected=presence_detected,
            presence_distance=self.presence_distance,
            extra_result=extra_result,
        )


@attrs.mutable(kw_only=True)
class _ProcessorConfig_v0(AlgoProcessorConfigBase):
    intra_enable: bool = attrs.field(default=True)
    inter_enable: bool = attrs.field(default=True)
    intra_detection_threshold: float = attrs.field(default=1.3)
    inter_detection_threshold: float = attrs.field(default=1)
    inter_phase_boost: bool = attrs.field(default=False)
    phase_adaptivity_tc: float = attrs.field(default=5)
    inter_frame_presence_timeout: Optional[int] = attrs.field(default=None)
    inter_frame_fast_cutoff: float = attrs.field(default=20.0)
    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    inter_output_time_const: float = attrs.field(default=5)
    intra_frame_time_const: float = attrs.field(default=0.15)
    intra_output_time_const: float = attrs.field(default=0.5)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        return []

    def migrate(self) -> ProcessorConfig:
        return ProcessorConfig(
            intra_enable=self.intra_enable,
            inter_enable=self.inter_enable,
            intra_detection_threshold=self.intra_detection_threshold,
            inter_detection_threshold=self.inter_detection_threshold,
            # phase_adaptivity_tc is removed
            inter_frame_presence_timeout=self.inter_frame_presence_timeout,
            inter_frame_fast_cutoff=self.inter_frame_fast_cutoff,
            inter_frame_slow_cutoff=self.inter_frame_slow_cutoff,
            inter_frame_deviation_time_const=self.inter_frame_deviation_time_const,
            inter_output_time_const=self.inter_output_time_const,
            intra_frame_time_const=self.intra_frame_time_const,
            intra_output_time_const=self.intra_output_time_const,
            # inter_phase_boost is removed
        )


processor_config_timeline = (
    tm.start(_ProcessorConfig_v0)
    .load(str, _ProcessorConfig_v0.from_json, fail=[TypeError])
    .load(dict, _ProcessorConfig_v0.from_dict, fail=[TypeError])
    .nop()
    .epoch(ProcessorConfig, _ProcessorConfig_v0.migrate, fail=[])
    .load(str, ProcessorConfig.from_json, fail=[TypeError])
    .load(dict, ProcessorConfig.from_dict, fail=[TypeError])
)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import typing as t

import h5py
import numpy as np

from acconeer.exptool._core.int_16_complex import INT_16_COMPLEX
from acconeer.exptool._core.recording import h5_record
from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities import (
    Metadata,
    Result,
    SensorCalibration,
    ServerInfo,
    SessionConfig,
)


def get_h5py_str_dtype() -> t.Any:
    return h5py.special_dtype(vlen=str)


_H5PY_STR_DTYPE = get_h5py_str_dtype()


class H5Saver(
    h5_record.H5Saver[
        SessionConfig,  # Config type
        t.List[t.Dict[int, Metadata]],  # Metadata type
        t.List[t.Dict[int, Result]],  # Result type
        ServerInfo,  # Server info type
    ]
):
    """H5Saver for A121 data"""

    _num_frames_current_session: int

    def __init__(self) -> None:
        self._num_frames_current_session = 0

    def _start(self) -> None:
        pass

    def _write_server_info(self, group: h5py.Group, server_info: ServerInfo) -> None:
        group.create_dataset(
            "server_info",
            data=server_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

    def _start_session(
        self,
        group: h5py.Group,
        config: SessionConfig,
        metadata: t.List[t.Dict[int, Metadata]],
        calibrations: t.Optional[t.Dict[int, SensorCalibration]] = None,
        calibrations_provided: t.Optional[t.Dict[int, bool]] = None,
        **kwargs: t.Any,
    ) -> None:
        group.create_dataset(
            "session_config",
            data=config.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

        for i, metadata_group_dict in enumerate(metadata):
            group_group = group.create_group(f"group_{i}")

            for entry_id, (sensor_id, single_metadata) in enumerate(metadata_group_dict.items()):
                entry_group = group_group.create_group(f"entry_{entry_id}")
                entry_group.create_dataset("sensor_id", data=sensor_id, track_times=False)

                entry_group.create_dataset(
                    "metadata",
                    data=single_metadata.to_json(),
                    dtype=_H5PY_STR_DTYPE,
                    track_times=False,
                )

                result_group = entry_group.create_group("result")
                self._create_result_datasets(result_group, single_metadata)

        if (calibrations is None) != (calibrations_provided is None):
            msg = "'calibrations_provided' must be provided if 'calibrations' is provided"
            raise ValueError(msg)

        if calibrations is not None and calibrations_provided is not None:
            calibrations_group = group.create_group("calibrations")
            for sensor_id, calibration in calibrations.items():
                sensor_calibration_group = calibrations_group.create_group(f"sensor_{sensor_id}")

                calibration.to_h5(sensor_calibration_group)

                sensor_calibration_group.create_dataset(
                    "provided", data=calibrations_provided[sensor_id], track_times=False
                )

    def _sample(self, group: h5py.Group, results: t.Iterable[t.List[t.Dict[int, Result]]]) -> None:
        self._num_frames_current_session += self._write_results_to_file(
            group=group,
            start_idx=self._num_frames_current_session,
            results=list(results),
        )

    @staticmethod
    def _create_result_datasets(g: h5py.Group, metadata: Metadata) -> None:
        g.create_dataset(
            "data_saturated",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "frame_delayed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "calibration_needed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "temperature",
            shape=(0,),
            maxshape=(None,),
            dtype=int,
            track_times=False,
            compression="gzip",
        )

        g.create_dataset(
            "tick",
            shape=(0,),
            maxshape=(None,),
            dtype=np.dtype("int64"),
            track_times=False,
            compression="gzip",
        )

        g.create_dataset(
            "frame",
            shape=(0, *metadata.frame_shape),
            maxshape=(None, *metadata.frame_shape),
            dtype=INT_16_COMPLEX,
            track_times=False,
            compression="gzip",
        )

    def _write_results_to_file(
        self, group: h5py.Group, start_idx: int, results: t.List[t.List[t.Dict[int, Result]]]
    ) -> int:
        """Saves the results to file.

        :returns: the number of extended results saved.
        """
        if len(results) == 0:
            return 0

        res: t.List[Result]
        for group_idx, entry_idx, res in utils.iterate_extended_structure_as_entry_list(
            utils.transpose_extended_structures(results)
        ):
            self._write_results(
                g=group[f"group_{group_idx}/entry_{entry_idx}/result"],
                start_index=start_idx,
                results=res,
            )

        return len(results)

    @staticmethod
    def _write_results(g: h5py.Group, start_index: int, results: list[Result]) -> None:
        """Extends the Dataset to the appropriate (new) size with .resize,
        and then copies the data over
        """
        datasets_to_extend = [
            "data_saturated",
            "frame_delayed",
            "calibration_needed",
            "temperature",
            "tick",
            "frame",
        ]
        for dataset_name in datasets_to_extend:
            g[dataset_name].resize(size=start_index + len(results), axis=0)

        dataset_slice = slice(start_index, start_index + len(results))

        g["data_saturated"][dataset_slice] = [result.data_saturated for result in results]
        g["frame_delayed"][dataset_slice] = [result.frame_delayed for result in results]
        g["calibration_needed"][dataset_slice] = [result.calibration_needed for result in results]
        g["temperature"][dataset_slice] = [result.temperature for result in results]
        g["tick"][dataset_slice] = [result.tick for result in results]
        g["frame"][dataset_slice] = [result._frame for result in results]

    def _stop_session(self, group: h5py.Group) -> None:
        self._num_frames_current_session = 0



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

from PySide6.QtCore import QByteArray, QPropertyAnimation, QSequentialAnimationGroup
from PySide6.QtGui import QColor
from PySide6.QtWidgets import (
    QGraphicsColorizeEffect,
    QWidget,
)

from acconeer.exptool.app.new.ui.icons import WARNING_YELLOW


def run_blink_animation(
    widget: QWidget,
    *,
    color: QColor = QColor(WARNING_YELLOW),
    num_blinks: int = 3,
    peak_strength: float = 0.8,
    rising_edge_ms: int = 100,
    falling_edge_ms: int = 300,
) -> None:
    c_effect = QGraphicsColorizeEffect(widget)
    c_effect.setColor(color)
    c_effect.setStrength(1.0)
    widget.setGraphicsEffect(c_effect)

    anim_group = QSequentialAnimationGroup(parent=widget)

    for _ in range(num_blinks):
        rising = QPropertyAnimation(c_effect, QByteArray(b"strength"))
        rising.setDuration(rising_edge_ms)
        rising.setStartValue(0.0)
        rising.setEndValue(peak_strength)
        anim_group.addAnimation(rising)

        falling = QPropertyAnimation(c_effect, QByteArray(b"strength"))
        falling.setDuration(falling_edge_ms)
        falling.setStartValue(peak_strength)
        falling.setEndValue(0.0)
        anim_group.addAnimation(falling)

    anim_group.start()



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget

from acconeer.exptool._core.entities.validation_result import Criticality, ValidationResult

from .group_box import GroupBox


class MiscErrorView(GroupBox[QVBoxLayout]):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(
            left_header="Miscellaneous Errors",
            layout_type=QVBoxLayout,
            parent=parent,
        )

        self.setVisible(False)

    def handle_validation_results(
        self, validation_results: list[ValidationResult]
    ) -> list[ValidationResult]:
        self.setVisible(len(validation_results) > 0)

        COLOR_MAP = {
            Criticality.ERROR: "#E6635A",
            Criticality.WARNING: "#FCC842",
            None: "white",
        }

        # Delete all existing error message widgets
        while self.layout().count():
            child = self.layout().takeAt(0)
            if child.widget():
                child.widget().deleteLater()

        for result in validation_results:
            error_widget = QLabel(parent=self)
            error_widget.setWordWrap(True)
            error_widget.setContentsMargins(5, 5, 5, 5)
            error_widget.setText(result.message)
            error_widget.setStyleSheet(
                f"background-color: {COLOR_MAP[result.criticality]}; "
                "color: white; font: bold italic;"
            )
            self.layout().addWidget(error_widget)

        return []



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import copy
import warnings
from enum import Enum
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool import type_migration as tm
from acconeer.exptool._core.class_creation.attrs import attrs_ndarray_isclose
from acconeer.exptool.a121._core.entities.configs.config_enums import IdleState, Profile
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._base import AlgoBase, AlgoBaseT
from acconeer.exptool.a121.algo._utils import estimate_frame_rate
from acconeer.exptool.a121.algo.presence._detector import (
    AlgoConfigBase,
    Controller,
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
)

from ._processor import Processor, ProcessorConfig, ProcessorResult


SPARSE_IQ_PPC = 24


def optional_profile_converter(profile: Optional[Profile]) -> Optional[Profile]:
    if profile is None:
        return None

    return Profile(profile)


def idle_state_converter(idle_state: IdleState) -> IdleState:
    return IdleState(idle_state)


@attributes_doc
@attrs.mutable(kw_only=True)
class PresenceZoneConfig(DetectorConfig):
    num_zones: int = attrs.field(default=7)
    """Maximum number of detection zones."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return super()._collect_validation_results()


@attributes_doc
@attrs.mutable(kw_only=True)
class PresenceWakeUpConfig(PresenceZoneConfig):
    num_zones_for_wake_up: int = attrs.field(default=1)
    """Number of detected zones needed for wake up."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results = super()._collect_validation_results()

        if self.num_zones_for_wake_up > self.num_zones:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "num_zones_for_wake_up",
                    f"Must not be more than number of zones ({self.num_zones})",
                )
            )

        return validation_results


@attributes_doc
@attrs.mutable(kw_only=True)
class RefAppConfig(AlgoConfigBase):
    nominal_config: PresenceZoneConfig = attrs.field(factory=PresenceZoneConfig)
    """Configuration used to keep detection."""

    wake_up_config: Optional[PresenceWakeUpConfig] = attrs.field(factory=PresenceWakeUpConfig)
    """Configuration used for wake up."""

    wake_up_mode: bool = attrs.field(default=True)
    """Switch between two configurations based on presence detection."""

    show_all_detected_zones: bool = attrs.field(default=False)
    """Visualize all detected zones in the circle sector."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        validation_results.extend(self.nominal_config._collect_validation_results())

        if self.wake_up_config is not None:
            validation_results.extend(self.wake_up_config._collect_validation_results())

        return validation_results

    @classmethod
    def from_dict(cls: type[AlgoBaseT], d: dict[str, Any]) -> AlgoBaseT:
        try:
            d["nominal_config"] = presence_zone_config_timeline.migrate(d["nominal_config"])
            d["wake_up_config"] = presence_wake_up_config_timeline.migrate(d["wake_up_config"])
        except tm.core.MigrationErrorGroup as exc:
            raise TypeError() from exc

        return cls(**d)


@attrs.mutable(kw_only=True)
class RefAppContext(AlgoBase):
    nominal_detector_context: Optional[DetectorContext] = attrs.field(default=None)
    wake_up_detector_context: Optional[DetectorContext] = attrs.field(default=None)

    @classmethod
    def from_dict(cls: type[AlgoBaseT], d: dict[str, Any]) -> AlgoBaseT:
        if d["nominal_detector_context"] is not None:
            d["nominal_detector_context"] = DetectorContext.from_dict(
                d["nominal_detector_context"]
            )
        if d["wake_up_detector_context"] is not None:
            d["wake_up_detector_context"] = DetectorContext.from_dict(
                d["wake_up_detector_context"]
            )

        return cls(**d)


@attrs.frozen(kw_only=True)
class RefAppResult:
    zone_limits: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    "The upper limit for each zone."

    presence_detected: bool = attrs.field()
    """True if presence was detected, False otherwise."""

    max_presence_zone: Optional[int] = attrs.field()
    """The zone for maximum presence score if presence detected, else None."""

    total_zone_detections: npt.NDArray[np.int_] = attrs.field(eq=attrs_ndarray_isclose)
    """Detection result for all zones."""

    inter_presence_score: float = attrs.field()
    """A measure of the amount of slow motion detected."""

    inter_zone_detections: npt.NDArray[np.int_] = attrs.field(eq=attrs_ndarray_isclose)
    """Slow motion presence detection result for all zones."""

    max_inter_zone: Optional[int] = attrs.field()
    """The zone for maximum slow motion presence score if slow presence detected, else None."""

    intra_presence_score: float = attrs.field()
    """A measure of the amount of fast motion detected."""

    intra_zone_detections: npt.NDArray[np.int_] = attrs.field(eq=attrs_ndarray_isclose)
    """Fast motion presence detection result for all zones."""

    max_intra_zone: Optional[int] = attrs.field()
    """The zone for maximum fast motion presence score if fast presence detected, else None."""

    used_config: _Mode = attrs.field()
    """The configuration used for the measurement."""

    wake_up_detections: Optional[npt.NDArray[np.int_]] = attrs.field()
    """Wake up detection result."""

    switch_delay: bool = attrs.field()
    """True if data was collected during switch delay."""

    service_result: a121.Result = attrs.field()


class _Mode(Enum):
    WAKE_UP_CONFIG = 0
    NOMINAL_CONFIG = 1


class RefApp(Controller[RefAppConfig, RefAppResult]):
    wake_up_detections: Optional[npt.NDArray[np.int_]]
    ref_app_context: RefAppContext

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        ref_app_config: RefAppConfig,
        ref_app_context: Optional[RefAppContext] = None,
    ) -> None:
        super().__init__(client=client, config=ref_app_config)
        self.sensor_id = sensor_id
        if ref_app_context is None:
            self.ref_app_context = RefAppContext()
        else:
            self.ref_app_context = ref_app_context

        self.started = False
        self.delay_count = 0
        if ref_app_config.wake_up_mode:
            assert ref_app_config.wake_up_config is not None
            self.wake_up_detections = np.zeros(ref_app_config.wake_up_config.num_zones, dtype=int)
            self.max_zone_time_n = int(np.around(2 * ref_app_config.wake_up_config.frame_rate))
        else:
            self.wake_up_detections = None

    def start(
        self, recorder: Optional[a121.Recorder] = None, algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        self.nominal_detector_config = self.config.nominal_config

        sensor_config = Detector._get_sensor_config(self.nominal_detector_config)
        session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )
        detector_config: DetectorConfig

        if self.ref_app_context.nominal_detector_context is None:
            self.nominal_detector_context = DetectorContext(
                estimated_frame_rate=estimate_frame_rate(self.client, session_config)
            )
            self.ref_app_context = RefAppContext(
                nominal_detector_context=self.nominal_detector_context
            )
        else:
            self.nominal_detector_context = self.ref_app_context.nominal_detector_context

        distances = np.linspace(
            self.config.nominal_config.start_m,
            self.config.nominal_config.end_m,
            sum([subsweep.num_points for subsweep in sensor_config.subsweeps]),
        )
        self.config.nominal_config.num_zones = min(
            self.config.nominal_config.num_zones, distances.size
        )

        self.nominal_processor_config = ProcessorConfig(
            num_zones=self.config.nominal_config.num_zones
        )

        if self.config.wake_up_mode:
            self._mode = _Mode.WAKE_UP_CONFIG
            assert self.config.wake_up_config is not None

            self.wake_up_detector_config = self.config.wake_up_config

            sensor_config = Detector._get_sensor_config(self.wake_up_detector_config)
            session_config = a121.SessionConfig(
                {self.sensor_id: sensor_config},
                extended=False,
            )

            if self.ref_app_context.wake_up_detector_context is None:
                self.wake_up_detector_context = DetectorContext(
                    estimated_frame_rate=estimate_frame_rate(self.client, session_config)
                )
                self.ref_app_context.wake_up_detector_context = self.wake_up_detector_context
            else:
                self.wake_up_detector_context = self.ref_app_context.wake_up_detector_context

            distances = np.linspace(
                self.config.wake_up_config.start_m,
                self.config.wake_up_config.end_m,
                sum([subsweep.num_points for subsweep in sensor_config.subsweeps]),
            )
            self.config.wake_up_config.num_zones = min(
                self.config.wake_up_config.num_zones, distances.size
            )
            self.config.wake_up_config.num_zones_for_wake_up = min(
                self.config.wake_up_config.num_zones_for_wake_up,
                self.config.wake_up_config.num_zones,
            )

            self.wake_up_processor_config = ProcessorConfig(
                num_zones=self.config.wake_up_config.num_zones
            )

            detector_config = self.wake_up_detector_config
            detector_context = self.wake_up_detector_context
            processor_config = self.wake_up_processor_config
        else:
            self._mode = _Mode.NOMINAL_CONFIG
            detector_config = self.nominal_detector_config
            detector_context = self.nominal_detector_context
            processor_config = self.nominal_processor_config

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("smart_presence")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.config,
                    self.ref_app_context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        self.detector = Detector(
            client=self.client,
            sensor_id=self.sensor_id,
            detector_config=detector_config,
            detector_context=detector_context,
        )
        self.detector.start(recorder=recorder, _algo_group=algo_group)
        assert self.detector.detector_metadata is not None
        session_config = self.detector.session_config

        self.ref_app_processor = Processor(
            processor_config,
            detector_config,
            session_config,
            self.detector.detector_metadata,
        )

        self.max_switch_delay_n = (
            np.maximum(
                (
                    self.config.nominal_config.inter_frame_deviation_time_const
                    + self.config.nominal_config.inter_output_time_const
                ),
                (
                    self.config.nominal_config.intra_frame_time_const
                    + self.config.nominal_config.intra_output_time_const
                ),
            )
            * self.config.nominal_config.frame_rate
        )

        self.started = True

    def get_next(self) -> RefAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.detector.get_next()
        processor_result = self.ref_app_processor.process(result)

        used_config = self._mode
        if self.config.wake_up_mode:
            self.determine_swapping(result, processor_result)

        return RefAppResult(
            zone_limits=processor_result.zone_limits,
            presence_detected=result.presence_detected,
            max_presence_zone=processor_result.max_presence_zone,
            total_zone_detections=processor_result.total_zone_detections,
            inter_presence_score=result.inter_presence_score,
            inter_zone_detections=processor_result.inter_zone_detections,
            max_inter_zone=processor_result.max_inter_zone,
            intra_presence_score=result.intra_presence_score,
            intra_zone_detections=processor_result.intra_zone_detections,
            max_intra_zone=processor_result.max_intra_zone,
            used_config=used_config,
            wake_up_detections=copy.deepcopy(self.wake_up_detections),
            switch_delay=self.delay_count > 0,
            service_result=result.service_result,
        )

    def determine_swapping(
        self, result: DetectorResult, processor_result: ProcessorResult
    ) -> None:
        if self.delay_count == 0:
            if self._mode == _Mode.WAKE_UP_CONFIG and result.presence_detected:
                assert self.config.wake_up_config is not None
                assert self.wake_up_detections is not None
                num_detections = 0
                for i, zone_detection in enumerate(processor_result.total_zone_detections):
                    if zone_detection == 1:
                        self.wake_up_detections[i] = self.max_zone_time_n

                    if self.wake_up_detections[i] > 0:
                        num_detections += 1
                        self.wake_up_detections[i] -= 1

                if num_detections >= self.config.wake_up_config.num_zones_for_wake_up:
                    self.swap_config(
                        self.nominal_detector_config,
                        self.nominal_processor_config,
                        self.nominal_detector_context,
                    )
                    self._mode = _Mode.NOMINAL_CONFIG
                    self.delay_count += 1
            elif self._mode == _Mode.NOMINAL_CONFIG and not result.presence_detected:
                self.swap_config(
                    self.wake_up_detector_config,
                    self.wake_up_processor_config,
                    self.wake_up_detector_context,
                )
                self._mode = _Mode.WAKE_UP_CONFIG
        else:
            if self.delay_count == 1:
                assert self.wake_up_detections is not None
                self.wake_up_detections.fill(0)

            self.delay_count += 1
            if self.delay_count >= self.max_switch_delay_n + 1 or result.presence_detected:
                self.delay_count = 0

    def swap_config(
        self,
        detector_config: DetectorConfig,
        processor_config: ProcessorConfig,
        detector_context: DetectorContext,
    ) -> None:
        self.detector.stop_detector()

        self.detector = Detector(
            client=self.client,
            sensor_id=self.sensor_id,
            detector_config=detector_config,
            detector_context=detector_context,
        )

        self.detector.start(recorder=None, _algo_group=None)
        assert self.detector.detector_metadata is not None

        session_config = self.detector.session_config
        processor_config = ProcessorConfig(num_zones=processor_config.num_zones)

        self.ref_app_processor = Processor(
            processor_config, detector_config, session_config, self.detector.detector_metadata
        )

    def update_config(self, config: RefAppConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        recorder_result = self.detector.stop()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    ref_app_config: RefAppConfig,
    ref_app_context: RefAppContext,
) -> None:
    algo_group.create_dataset(
        "ref_app_sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "ref_app_config", ref_app_config.to_json())
    _create_h5_string_dataset(algo_group, "ref_app_context", ref_app_context.to_json())


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, RefAppConfig, RefAppContext]:
    sensor_id = int(algo_group["ref_app_sensor_id"][()])
    config = RefAppConfig.from_json(algo_group["ref_app_config"][()])
    ref_app_context = RefAppContext.from_json(algo_group["ref_app_context"][()])
    return sensor_id, config, ref_app_context


@attrs.mutable(kw_only=True)
class _PresenceZoneConfig_v0(AlgoConfigBase):
    num_zones: int = attrs.field(default=7)
    start_m: float = attrs.field(default=0.3)
    end_m: float = attrs.field(default=2.5)
    profile: Optional[a121.Profile] = attrs.field(
        default=None, converter=optional_profile_converter
    )
    step_length: Optional[int] = attrs.field(default=None)
    frame_rate: float = attrs.field(default=12.0)
    sweeps_per_frame: int = attrs.field(default=16)
    automatic_subsweeps: bool = attrs.field(default=False)
    signal_quality: float = attrs.field(default=15.0)
    hwaas: int = attrs.field(default=32)
    inter_frame_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.DEEP_SLEEP, converter=idle_state_converter
    )
    intra_enable: bool = attrs.field(default=True)
    intra_detection_threshold: float = attrs.field(default=1.3)
    intra_frame_time_const: float = attrs.field(default=0.15)
    intra_output_time_const: float = attrs.field(default=0.3)
    inter_enable: bool = attrs.field(default=True)
    inter_detection_threshold: float = attrs.field(default=1)
    inter_frame_fast_cutoff: float = attrs.field(default=6.0)
    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    inter_output_time_const: float = attrs.field(default=2)
    inter_frame_presence_timeout: Optional[int] = attrs.field(default=3)
    inter_phase_boost: bool = attrs.field(default=False)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return []

    def migrate(self) -> PresenceZoneConfig:
        return PresenceZoneConfig(
            num_zones=self.num_zones,
            start_m=self.start_m,
            end_m=self.end_m,
            profile=self.profile,
            step_length=self.step_length,
            frame_rate=self.frame_rate,
            sweeps_per_frame=self.sweeps_per_frame,
            automatic_subsweeps=self.automatic_subsweeps,
            signal_quality=self.signal_quality,
            hwaas=self.hwaas,
            inter_frame_idle_state=self.inter_frame_idle_state,
            intra_enable=self.intra_enable,
            intra_detection_threshold=self.intra_detection_threshold,
            intra_frame_time_const=self.intra_frame_time_const,
            intra_output_time_const=self.intra_output_time_const,
            inter_enable=self.inter_enable,
            inter_detection_threshold=self.inter_detection_threshold,
            inter_frame_fast_cutoff=self.inter_frame_fast_cutoff,
            inter_frame_slow_cutoff=self.inter_frame_slow_cutoff,
            inter_frame_deviation_time_const=self.inter_frame_deviation_time_const,
            inter_output_time_const=self.inter_output_time_const,
            inter_frame_presence_timeout=self.inter_frame_presence_timeout,
            # inter_phase_boost is removed
        )


@attrs.mutable(kw_only=True)
class _PresenceWakeUpConfig_v0(AlgoConfigBase):
    num_zones_for_wake_up: int = attrs.field(default=1)
    num_zones: int = attrs.field(default=7)
    start_m: float = attrs.field(default=0.3)
    end_m: float = attrs.field(default=2.5)
    profile: Optional[a121.Profile] = attrs.field(
        default=None, converter=optional_profile_converter
    )
    step_length: Optional[int] = attrs.field(default=None)
    frame_rate: float = attrs.field(default=12.0)
    sweeps_per_frame: int = attrs.field(default=16)
    automatic_subsweeps: bool = attrs.field(default=False)
    signal_quality: float = attrs.field(default=15.0)
    hwaas: int = attrs.field(default=32)
    inter_frame_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.DEEP_SLEEP, converter=idle_state_converter
    )
    intra_enable: bool = attrs.field(default=True)
    intra_detection_threshold: float = attrs.field(default=1.3)
    intra_frame_time_const: float = attrs.field(default=0.15)
    intra_output_time_const: float = attrs.field(default=0.3)
    inter_enable: bool = attrs.field(default=True)
    inter_detection_threshold: float = attrs.field(default=1)
    inter_frame_fast_cutoff: float = attrs.field(default=6.0)
    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    inter_output_time_const: float = attrs.field(default=2)
    inter_frame_presence_timeout: Optional[int] = attrs.field(default=3)
    inter_phase_boost: bool = attrs.field(default=False)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return []

    def migrate(self) -> PresenceWakeUpConfig:
        return PresenceWakeUpConfig(
            num_zones_for_wake_up=self.num_zones_for_wake_up,
            num_zones=self.num_zones,
            start_m=self.start_m,
            end_m=self.end_m,
            profile=self.profile,
            step_length=self.step_length,
            frame_rate=self.frame_rate,
            sweeps_per_frame=self.sweeps_per_frame,
            automatic_subsweeps=self.automatic_subsweeps,
            signal_quality=self.signal_quality,
            hwaas=self.hwaas,
            inter_frame_idle_state=self.inter_frame_idle_state,
            intra_enable=self.intra_enable,
            intra_detection_threshold=self.intra_detection_threshold,
            intra_frame_time_const=self.intra_frame_time_const,
            intra_output_time_const=self.intra_output_time_const,
            inter_enable=self.inter_enable,
            inter_detection_threshold=self.inter_detection_threshold,
            inter_frame_fast_cutoff=self.inter_frame_fast_cutoff,
            inter_frame_slow_cutoff=self.inter_frame_slow_cutoff,
            inter_frame_deviation_time_const=self.inter_frame_deviation_time_const,
            inter_output_time_const=self.inter_output_time_const,
            inter_frame_presence_timeout=self.inter_frame_presence_timeout,
            # inter_phase_boost is removed
        )


presence_zone_config_timeline = (
    tm.start(_PresenceZoneConfig_v0)
    .load(str, _PresenceZoneConfig_v0.from_json, fail=[TypeError])
    .load(dict, _PresenceZoneConfig_v0.from_dict, fail=[TypeError])
    .nop()
    .epoch(PresenceZoneConfig, _PresenceZoneConfig_v0.migrate, fail=[])
    .load(str, PresenceZoneConfig.from_json, fail=[TypeError])
    .load(dict, PresenceZoneConfig.from_dict, fail=[TypeError])
)


presence_wake_up_config_timeline = (
    tm.start(_PresenceWakeUpConfig_v0)
    .load(str, _PresenceWakeUpConfig_v0.from_json, fail=[TypeError])
    .load(dict, _PresenceWakeUpConfig_v0.from_dict, fail=[TypeError])
    .nop()
    .epoch(PresenceWakeUpConfig, _PresenceWakeUpConfig_v0.migrate, fail=[])
    .load(str, PresenceWakeUpConfig.from_json, fail=[TypeError])
    .load(dict, PresenceWakeUpConfig.from_dict, fail=[TypeError])
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import contextlib
import functools
import json
import logging
import queue
import shutil
import time
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from uuid import UUID

import attrs
from packaging.version import Version
from typing_extensions import Protocol

from PySide6.QtCore import QDeadlineTimer, QObject, QThread, Signal
from PySide6.QtWidgets import QApplication, QWidget

from acconeer.exptool import _core as core
from acconeer.exptool import a121
from acconeer.exptool._core.communication.comm_devices import CommDevice, SerialDevice, USBDevice
from acconeer.exptool.app.new._enums import (
    ConnectionInterface,
    ConnectionState,
    PluginGeneration,
    PluginState,
)
from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.app.new.app_model.file_detective import investigate_file
from acconeer.exptool.app.new.backend import (
    Backend,
    BackendPlugin,
    BackendPluginStateMessage,
    ClosedTask,
    ConnectionStateMessage,
    GeneralMessage,
    LogMessage,
    Message,
    Model,
    PlotMessage,
    PluginStateMessage,
    StatusFileAccessMessage,
    StatusMessage,
    Task,
    TimingMessage,
    _RateStats,
)
from acconeer.exptool.app.new.storage import get_config_dir, remove_temp_dir

from .plugin_protocols import PlotPluginInterface
from .port_updater import PortUpdater


_AnyClient = core.Client[Any, Any, Any, Any, Any]
log = logging.getLogger(__name__)


class PluginPresetSpec(Protocol):
    """Defines what AppModel needs to know about a plugin preset.

    Implementations are free to add additional fields.
    """

    name: str
    description: Optional[str]
    preset_id: Enum


class PluginSpec(Protocol):
    """Defines what AppModel needs to know about a plugin.

    Implementations are free to add additional fields.
    """

    key: str
    generation: PluginGeneration
    presets: List[PluginPresetSpec]
    title: str
    default_preset_id: Enum

    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin[Any]: ...

    def create_view_plugin(self, app_model: AppModel) -> QWidget: ...

    def create_plot_plugin(self, app_model: AppModel) -> PlotPluginInterface: ...


class _BackendListeningThread(QThread):
    sig_backend_closed_task = Signal(ClosedTask)
    sig_backend_message = Signal(Message)

    def __init__(self, backend: Backend, parent: QObject) -> None:
        super().__init__(parent)
        self.backend = backend

    def run(self) -> None:
        log.debug("Backend listening thread starting...")

        while not self.isInterruptionRequested():
            try:
                item = self.backend.recv(timeout=0.1)
            except queue.Empty:
                continue

            if isinstance(item, Message):
                self.sig_backend_message.emit(item)
            elif isinstance(item, ClosedTask):
                self.sig_backend_closed_task.emit(item)
            else:
                raise AssertionError

        log.debug("Backend listening thread stopping...")


def _to_usbdevice(obj: Union[dict[str, Any], USBDevice]) -> Optional[USBDevice]:
    if isinstance(obj, dict):
        return USBDevice.from_dict(obj)
    return obj


def _to_serialdevice(obj: Union[dict[str, Any], SerialDevice]) -> Optional[SerialDevice]:
    if isinstance(obj, dict):
        return SerialDevice.from_dict(obj)
    return obj


@attrs.mutable(kw_only=True)
class _PersistentState:
    _FILE_NAME = "app_model_state.json"
    _ENUMS = {
        "ConnectionInterface": ConnectionInterface,
        "PluginGeneration": PluginGeneration,
    }

    class Encoder(json.JSONEncoder):
        def default(self, obj: Any) -> Any:
            if type(obj) in _PersistentState._ENUMS.values():
                # An enum existing in _ENUMS will be transformed to:
                # {"__enum__": "RegisteredEnum.MEMBER"}
                return {"__enum__": str(obj)}
            return json.JSONEncoder.default(self, obj)

    @staticmethod
    def _to_enum(d: dict[str, Any]) -> Any:
        if "__enum__" in d:
            # Transform the dict {"__enum__": "RegisteredEnum.MEMBER"}
            # to the enum RegisteredEnum.MEMBER
            name, member = d["__enum__"].split(".")
            return getattr(_PersistentState._ENUMS[name], member)
        return d

    connection_interface: ConnectionInterface = attrs.field(default=ConnectionInterface.SERIAL)
    plugin_generation: PluginGeneration = attrs.field(default=PluginGeneration.A121)
    socket_connection_ip: str = attrs.field(default="")
    serial_connection_device: Optional[SerialDevice] = attrs.field(
        default=None, converter=_to_serialdevice
    )
    usb_connection_device: Optional[USBDevice] = attrs.field(default=None, converter=_to_usbdevice)
    overridden_baudrate: Optional[int] = attrs.field(default=None)
    autoconnect_enabled: bool = attrs.field(default=False)
    flow_control_enabled: bool = attrs.field(default=True)
    recording_enabled: bool = attrs.field(default=True)

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> _PersistentState:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), cls=self.Encoder)

    @classmethod
    def from_json(cls, json_str: str) -> _PersistentState:
        return cls.from_dict(json.loads(json_str, object_hook=_PersistentState._to_enum))

    @classmethod
    def from_config_file(cls) -> _PersistentState:
        return cls.from_json((get_config_dir() / cls._FILE_NAME).read_text(encoding="utf-8"))

    def to_config_file(self) -> None:
        (get_config_dir() / _PersistentState._FILE_NAME).write_text(
            self.to_json(), encoding="utf-8"
        )


class AppModel(QObject):
    sig_notify = Signal(object)
    sig_error = Signal(Exception, object)
    sig_load_plugin = Signal(object)
    sig_message_plot_plugin = Signal(PlotMessage)
    sig_message_view_plugin = Signal(object)
    sig_status_file_path = Signal(str, bool)
    sig_status_message = Signal(object)
    sig_rate_stats = Signal(float, bool, float, bool)
    sig_backend_cpu_percent = Signal(int)
    sig_frame_count = Signal(object)
    sig_backend_state_changed = Signal(object)
    sig_resource_tab_input_block_requested = Signal(object)
    """Emit a config object to spawn an input block in the resource tab"""
    sig_timing = Signal(tuple)
    """Signals the elapsed time of different things: tuple[<name>, <start time (s)>, <end time (s)>]"""

    plugins: list[PluginSpec]
    plugin: Optional[PluginSpec]

    backend_plugin_state: Any

    connection_warning: Optional[str]
    available_serial_devices: List[SerialDevice]
    available_usb_devices: List[USBDevice]

    saveable_file: Optional[Path]

    def __init__(
        self,
        backend: Backend,
        plugins: list[PluginSpec],
        force_autoconnect: bool = False,
        socket_autoconnect: Optional[str] = None,
        simulated_autocconnect: bool = False,
    ) -> None:
        super().__init__()
        self._backend = backend
        self._listener = _BackendListeningThread(self._backend, self)
        self._listener.sig_backend_message.connect(self._handle_backend_message)
        self._listener.sig_backend_closed_task.connect(self._handle_backend_closed_task)
        self._port_updater = PortUpdater(self)
        self._port_updater.sig_update.connect(self._handle_port_update)

        self._backend_task_callbacks: dict[UUID, Any] = {}
        self._backend_task_send_timestamps: dict[UUID, float] = {}
        self._force_autoconnect = force_autoconnect

        self._a121_server_info: Optional[a121.ServerInfo] = None

        try:
            self._persistent_state = _PersistentState.from_config_file()
        except Exception as exc:
            if not isinstance(exc, FileNotFoundError):
                log.error("Config file loading failed, using defaults")
            self._persistent_state = _PersistentState()

        self.plugins = plugins
        self.plugin = None

        self.backend_plugin_state = None

        self._connection_state = ConnectionState.DISCONNECTED
        self.connection_warning = None
        self._plugin_state = PluginState.UNLOADED
        self._serial_connection_device: Optional[SerialDevice] = None
        self._usb_connection_device: Optional[USBDevice] = None
        self.available_serial_devices = []
        self.available_usb_devices = []

        self.saveable_file = None

        if socket_autoconnect is not None:
            self.set_connection_interface(ConnectionInterface.SOCKET)
            self.set_socket_connection_ip(socket_autoconnect)
            self.connect_client(auto=True)

        if simulated_autocconnect:
            self.set_connection_interface(ConnectionInterface.SIMULATED)
            self.connect_client(auto=True)

    @contextlib.contextmanager
    def report_timing(self, name: str) -> Iterator[None]:
        start = time.perf_counter()
        yield
        end = time.perf_counter()
        self.sig_timing.emit((name, start, end))

    @property
    def plugin_state(self) -> PluginState:
        """Read-only property of the plugin state"""
        return self._plugin_state

    @property
    def connection_state(self) -> ConnectionState:
        """Read-only property of the connection state"""
        return self._connection_state

    def start(self) -> None:
        self._listener.start()
        self.port_updater_enable(True)

    def stop(self) -> None:
        WAIT_FOR_UNLOAD_TIMEOUT = 1.0

        try:
            self._persistent_state.to_config_file()
        except Exception:
            log.error("Config saving failed")

        self.load_plugin(None)
        if self.connection_state in [ConnectionState.CONNECTING, ConnectionState.CONNECTED]:
            self.disconnect_client()

        wait_start_time = time.monotonic()
        while (
            self.plugin_state != PluginState.UNLOADED
            and self.connection_state != ConnectionState.DISCONNECTED
        ):  # TODO: Do this better
            QApplication.processEvents()

            if (time.monotonic() - wait_start_time) > WAIT_FOR_UNLOAD_TIMEOUT:
                log.error("Plugin not unloaded on stop")
                break

        remove_temp_dir()

        self._listener.requestInterruption()
        status = self._listener.wait(QDeadlineTimer(500))

        if not status:
            log.debug("Backend listening thread did not stop when requested, terminating...")
            self._listener.terminate()

        self.port_updater_enable(False)

    def broadcast(self) -> None:
        self.sig_notify.emit(self)

    def broadcast_backend_state(self) -> None:
        self.sig_backend_state_changed.emit(self.backend_plugin_state)

    def emit_error(self, exception: Exception, traceback_format_exc: Optional[str] = None) -> None:
        log.debug("Emitting error")
        self.sig_error.emit(exception, traceback_format_exc)

    def put_task(
        self,
        task: Task,
        *,
        on_ok: Optional[Callable[[], None]] = None,
        on_error: Optional[Callable[[Exception, Optional[str]], None]] = None,
    ) -> None:
        key = self._backend.put_task(task)
        self._backend_task_callbacks[key] = {
            "on_ok": on_ok,
            "on_error": on_error or self.emit_error,
        }
        self._backend_task_send_timestamps[key] = time.perf_counter()

        (name, _) = task
        log.debug(f"Put backend task with name: '{name}', key: {key.time_low}")

    def _handle_backend_closed_task(self, closed_task: ClosedTask) -> None:
        log.debug(f"Got backend closed task: {closed_task.key.time_low}")

        callbacks = self._backend_task_callbacks.pop(closed_task.key)
        task_send_timestamp = self._backend_task_send_timestamps.pop(closed_task.key)
        task_roundtrip = ("Task Roundtrip", task_send_timestamp, time.perf_counter())
        self.sig_timing.emit(task_roundtrip)

        if closed_task.exception:
            f = callbacks["on_error"]
            if f:
                f(closed_task.exception, closed_task.traceback_format_exc)
        else:
            f = callbacks["on_ok"]
            if f:
                f()

    def _handle_backend_message(self, message: Message) -> None:
        if isinstance(message, ConnectionStateMessage):
            log.debug(f"Got backend connection state message {message.state}")
            self.set_connection_state(message.state)
            self.connection_warning = message.warning
            self.broadcast()
        elif isinstance(message, PluginStateMessage):
            log.debug(f"Got plugin state message {message.state}")
            self._plugin_state = message.state
            self.broadcast()
        elif isinstance(message, BackendPluginStateMessage):
            log.debug("Got backend plugin state message")
            self.backend_plugin_state = message.state
            self.broadcast_backend_state()
            self.broadcast()
        elif isinstance(message, StatusFileAccessMessage):
            self.send_status_file_path(message.file_path, message.opened)
        elif isinstance(message, StatusMessage):
            self.send_status_message(message.status)
        elif isinstance(message, LogMessage):
            module_logger = logging.getLogger(message.module_name)
            loglevel_to_logfunc = {
                "CRITICAL": module_logger.critical,
                "ERROR": module_logger.error,
                "WARNING": module_logger.warning,
                "INFO": module_logger.info,
                "DEBUG": module_logger.debug,
            }
            loglevel_to_logfunc[message.log_level](message.log_string)
        elif isinstance(message, GeneralMessage):
            if message.recipient is not None:
                if message.recipient == "plot_plugin":
                    self.sig_message_plot_plugin.emit(message)
                elif message.recipient == "view_plugin":
                    self.sig_message_view_plugin.emit(message)
                else:
                    msg = f"Got message with unknown recipient '{message.recipient}'"
                    raise RuntimeError(msg)
            else:
                self._handle_backend_general_message(message)
        elif isinstance(message, TimingMessage):
            self.sig_timing.emit((message.name, message.start, message.end))
        else:
            msg = f"Got message of unknown type '{type(message)}'"
            raise RuntimeError(msg)

    def _handle_backend_general_message(self, message: GeneralMessage) -> None:
        if message.exception:
            self.emit_error(message.exception, message.traceback_format_exc)
            return

        if message.name == "server_info":
            server_info = message.data
            if isinstance(server_info, a121.ServerInfo):
                self._a121_server_info = server_info
            else:
                msg = f"Unexpected server_info type: {type(server_info)}"
                raise TypeError(msg)

            self.broadcast()
        elif message.name == "saveable_file":
            assert message.data is None or isinstance(message.data, Path)
            self._update_saveable_file(message.data)
        elif message.name == "rate_stats":
            stats = message.data
            if stats is None:
                stats = _RateStats.invalid()
            else:
                assert isinstance(stats, _RateStats)

            self.sig_rate_stats.emit(
                stats.rate,
                stats.rate_warning,
                stats.jitter,
                stats.jitter_warning,
            )
        elif message.name == "cpu_percent":
            self.sig_backend_cpu_percent.emit(message.data)
        elif message.name == "frame_count":
            self.sig_frame_count.emit(message.data)
        else:
            msg = f"Got unknown general message '{message.name}'"
            raise RuntimeError(msg)

    def _update_saveable_file(self, path: Optional[Path]) -> None:
        if self.saveable_file is not None:
            self.saveable_file.unlink(missing_ok=True)

        self.saveable_file = path
        self.broadcast()

    def _is_serial_device_unflashed(self, serial_device: Optional[SerialDevice]) -> bool:
        return bool(serial_device and serial_device.unflashed)

    def _is_usb_device_unflashed(self, usb_device: Optional[USBDevice]) -> bool:
        return bool(usb_device and usb_device.unflashed)

    def _is_usb_device_inaccessible(self, usb_device: Optional[USBDevice]) -> bool:
        return bool(usb_device and not usb_device.accessible)

    def _handle_port_update(
        self,
        serial_devices: list[SerialDevice],
        usb_devices: list[USBDevice],
    ) -> None:
        if self.connection_state is not ConnectionState.DISCONNECTED and (
            (
                self.connection_interface == ConnectionInterface.SERIAL
                and self.serial_connection_device not in serial_devices
            )
            or (
                self.connection_interface == ConnectionInterface.USB
                and self.usb_connection_device not in usb_devices
            )
        ):
            self.disconnect_client()

        first_update = len(self.available_usb_devices + self.available_serial_devices) == 0 and (
            self._persistent_state.usb_connection_device
            or self._persistent_state.serial_connection_device
        )

        serial_connection_device, recognized = self._select_new_device(
            self.available_serial_devices,
            serial_devices,
            self._persistent_state.serial_connection_device,
        )
        if serial_connection_device is None or isinstance(serial_connection_device, SerialDevice):
            self.set_serial_connection_device(serial_connection_device)
            self.available_serial_devices = serial_devices

        connect = False

        if recognized:
            self.set_connection_interface(ConnectionInterface.SERIAL)
            if self._is_serial_device_unflashed(self.serial_connection_device):
                connect = False
                self.send_warning_message(
                    f"Found unflashed device at serial port: {self.serial_connection_device}"
                )
            else:
                connect = True
                self.send_status_message(
                    f"Recognized serial port: {self.serial_connection_device}"
                )

        usb_connection_device, recognized = self._select_new_device(
            self.available_usb_devices,
            usb_devices,
            self._persistent_state.usb_connection_device,
        )
        if usb_connection_device is None or isinstance(usb_connection_device, USBDevice):
            self.set_usb_connection_device(usb_connection_device)
            self.available_usb_devices = usb_devices

        if recognized:
            assert usb_connection_device is not None
            assert isinstance(usb_connection_device, USBDevice)
            self.set_connection_interface(ConnectionInterface.USB)
            if self._is_usb_device_unflashed(usb_connection_device):
                connect = False
                self.send_warning_message(f"Found unflashed USB device: {usb_connection_device}")
            elif self._is_usb_device_inaccessible(usb_connection_device):
                connect = False
                self.send_warning_message(
                    f"Found inaccessible USB device: {usb_connection_device}"
                )
            else:
                connect = True
                self.send_status_message(f"Recognized USB device: {usb_connection_device}")

        if (
            (connect or first_update)
            and self.connection_state == ConnectionState.DISCONNECTED
            and self.autoconnect_enabled
        ):
            self._autoconnect()

        self.broadcast()

    def _autoconnect(self) -> None:
        self.connect_client(auto=True)

    def _select_new_device(
        self,
        old_devices: Sequence[CommDevice],
        new_devices: Sequence[CommDevice],
        current_port: Optional[CommDevice],
    ) -> Tuple[Optional[CommDevice], bool]:
        if self.connection_state != ConnectionState.DISCONNECTED:
            return current_port, False

        if not new_devices:
            return None, False

        added_devices = [dev for dev in new_devices if dev not in old_devices]

        # This can happen in the first port update when the current device is
        # restored from persistent state
        if current_port in added_devices:
            return current_port, False

        for device in added_devices:
            if device.recognized:
                return device, True

        if current_port not in new_devices:
            return new_devices[0], True

        return current_port, False

    def connect_client(self, auto: bool = False) -> None:
        open_client_parameters: Dict[str, Any]

        if self.connection_interface == ConnectionInterface.SIMULATED:
            open_client_parameters = {"mock": True}
        elif self.connection_interface == ConnectionInterface.SOCKET:
            open_client_parameters = {"ip_address": self.socket_connection_ip}
        elif (
            self.connection_interface == ConnectionInterface.SERIAL
            and self.serial_connection_device is not None
        ):
            open_client_parameters = {
                "serial_port": self.serial_connection_device.port,
                "override_baudrate": self.overridden_baudrate,
                "flow_control": self.flow_control_enabled,
            }
        elif (
            self.connection_interface == ConnectionInterface.USB
            and self.usb_connection_device is not None
        ):
            if self.usb_connection_device.serial is not None:
                open_client_parameters = {"usb_device": self.usb_connection_device.serial}
            else:
                open_client_parameters = {"usb_device": True}
        else:
            raise RuntimeError

        log.debug(f"Connecting client with {open_client_parameters}")

        get_connection_warning = functools.partial(
            self._connection_warning, Version(a121.SDK_VERSION)
        )
        task_putter = functools.partial(
            self.put_task,
            on_error=self._failed_autoconnect if auto else self.emit_error,
        )

        Model.connect_client.rpc(
            task_putter,
            client_factory=self._create_client_factory(
                self.plugin_generation, open_client_parameters
            ),
            get_connection_warning=get_connection_warning,
        )
        self.connection_warning = None
        self.broadcast()

    @staticmethod
    def _create_client_factory(
        generation: PluginGeneration, open_kwargs: dict[str, Any]
    ) -> Callable[[], _AnyClient]:
        factory: Callable[..., _AnyClient]
        if generation == PluginGeneration.A121:
            factory = a121.Client.open
        else:
            msg = f"Don't know how to construct a client for generation {generation}"
            raise NotImplementedError(msg)

        return functools.partial(factory, **open_kwargs)

    @staticmethod
    def _connection_warning(supported: Version, server: Version) -> Optional[str]:
        if server > supported:
            return "New server version - please upgrade client"
        elif server < supported:
            return "Old server version - please upgrade server"
        else:
            return None

    def disconnect_client(self) -> None:
        Model.disconnect_client.rpc(self.put_task)
        self.connection_warning = None
        self._a121_server_info = None
        self.broadcast()

    def is_connect_ready(self) -> bool:
        return (
            (self.connection_interface == ConnectionInterface.SOCKET)
            or (self.connection_interface == ConnectionInterface.SIMULATED)
            or (
                self.connection_interface == ConnectionInterface.SERIAL
                and self.serial_connection_device is not None
                and not self._is_serial_device_unflashed(self.serial_connection_device)
            )
            or (
                self.connection_interface == ConnectionInterface.USB
                and self.usb_connection_device is not None
                and not self._is_usb_device_unflashed(self.usb_connection_device)
                and not self._is_usb_device_inaccessible(self.usb_connection_device)
            )
        )

    def _failed_autoconnect(
        self, exception: Exception, traceback_format_exc: Optional[str] = None
    ) -> None:
        self.send_warning_message("Failed to autoconnect")

    def port_updater_enable(self, enable: bool) -> None:
        if enable:
            self._port_updater.start()
        else:
            self._port_updater.stop()

    def set_connection_state(self, new_connection_state: ConnectionState) -> None:
        if new_connection_state == ConnectionState.DISCONNECTED:
            self.port_updater_enable(True)
        else:
            self.port_updater_enable(False)
        self._connection_state = new_connection_state

    @property
    def connection_interface(self) -> ConnectionInterface:
        return self._persistent_state.connection_interface

    @property
    def socket_connection_ip(self) -> str:
        return self._persistent_state.socket_connection_ip

    @property
    def serial_connection_device(self) -> Optional[SerialDevice]:
        return self._serial_connection_device

    @property
    def usb_connection_device(self) -> Optional[USBDevice]:
        return self._usb_connection_device

    @property
    def autoconnect_enabled(self) -> bool:
        return self.autoconnect_forced or self._persistent_state.autoconnect_enabled

    @property
    def autoconnect_forced(self) -> bool:
        return self._force_autoconnect

    @property
    def flow_control_enabled(self) -> bool:
        return self._persistent_state.flow_control_enabled

    @property
    def overridden_baudrate(self) -> Optional[int]:
        return self._persistent_state.overridden_baudrate

    @property
    def recording_enabled(self) -> bool:
        return self._persistent_state.recording_enabled

    @property
    def plugin_generation(self) -> PluginGeneration:
        return self._persistent_state.plugin_generation

    def set_connection_interface(self, connection_interface: ConnectionInterface) -> None:
        self._persistent_state.connection_interface = connection_interface
        self.broadcast()

    def set_socket_connection_ip(self, ip: str) -> None:
        self._persistent_state.socket_connection_ip = ip
        self.broadcast()

    def set_serial_connection_device(self, device: Optional[SerialDevice]) -> None:
        self._serial_connection_device = device
        self._persistent_state.serial_connection_device = device
        self.broadcast()

    def set_usb_connection_device(self, device: Optional[USBDevice]) -> None:
        self._usb_connection_device = device
        self._persistent_state.usb_connection_device = device
        self.broadcast()

    def set_autoconnect_enabled(self, autoconnect_enabled: bool) -> None:
        self._persistent_state.autoconnect_enabled = autoconnect_enabled
        self.broadcast()

    def set_flow_control_enabled(self, flow_control_enabled: bool) -> None:
        self._persistent_state.flow_control_enabled = flow_control_enabled
        self.broadcast()

    def set_overridden_baudrate(self, overridden_baudrate: Optional[int]) -> None:
        self._persistent_state.overridden_baudrate = overridden_baudrate
        self.broadcast()

    def set_recording_enabled(self, recording_enabled: bool) -> None:
        self._persistent_state.recording_enabled = recording_enabled
        self.broadcast()

    def set_plugin_generation(self, new_generation: PluginGeneration) -> None:
        self._persistent_state.plugin_generation = new_generation
        self.broadcast()

    def _unload_current_plugin(self) -> None:
        log.debug("AppModel is unloading its current plugin")
        self.sig_load_plugin.emit(None)
        self._update_saveable_file(None)
        self.backend_plugin_state = None
        self.broadcast()

        Model.unload_plugin.rpc(self.put_task)

    def load_plugin(self, plugin: Optional[PluginSpec]) -> None:
        log.debug(f"AppModel is loading the plugin {plugin}")
        if plugin == self.plugin:
            return

        self._unload_current_plugin()

        if plugin is not None:
            Model.load_plugin.rpc(
                self.put_task,
                plugin_factory=plugin.create_backend_plugin,
                key=plugin.key,
            )
            BackendPlugin.load_from_cache.rpc(self.put_task)

        self.sig_load_plugin.emit(plugin)
        self.plugin = plugin
        self.broadcast()
        self.broadcast_backend_state()

    def set_plugin_preset(self, preset_id: Enum) -> None:
        BackendPlugin.set_preset.rpc(self.put_task, preset_id=preset_id.value)

    def save_to_file(self, path: Path) -> None:
        log.debug(f"{self.__class__.__name__} saving to file '{path}'")

        if self.saveable_file is None:
            raise RuntimeError

        shutil.copyfile(self.saveable_file, path)

    def load_from_file(self, path: Path) -> None:
        log.debug(f"{self.__class__.__name__} loading from file '{path}'")

        findings = investigate_file(path)

        if findings is None:
            self.emit_error(HandledException("Cannot load file"))
            return

        if findings.generation != PluginGeneration.A121:
            self.emit_error(HandledException("This app can currently only load A121 files"))
            return

        if findings.key is None:
            plugin = self.find_plugin("sparse_iq", self.plugin_generation)  # noqa: F841
        else:
            plugin = self.find_plugin(findings.key, self.plugin_generation)

            if plugin is None:
                log.debug(f"Could not find plugin '{findings.key}'")
                return

        self.load_plugin(plugin)
        BackendPlugin.load_from_file.rpc(self.put_task, path=path)

    def find_plugin(self, plugin_key: str, generation: PluginGeneration) -> Optional[PluginSpec]:
        return next(
            (
                plugin
                for plugin in self.plugins
                if plugin.key == plugin_key and plugin.generation == generation
            ),
            None,
        )

    @property
    def rss_version(self) -> Optional[str]:
        if self._a121_server_info is not None:
            return self._a121_server_info.rss_version

        return None

    @property
    def connected_sensors(self) -> list[int]:
        return self._a121_server_info.connected_sensors if self._a121_server_info else []

    def send_status_file_path(self, file_path: str, opened: bool) -> None:
        self.sig_status_file_path.emit(file_path, opened)

    def send_status_message(self, message: Optional[str]) -> None:
        self.sig_status_message.emit(message)

    def send_warning_message(self, message: Optional[str]) -> None:
        self.sig_status_message.emit(f'<p style="color: #FD5200;"><b>{message}</b></p>')



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from .h5_record import (
    _H5PY_STR_DTYPE,
    H5Record,
    H5Recorder,
    RecordError,
    load_record,
    open_record,
    save_record,
    save_record_to_h5,
)
from .im_record import InMemoryRecord
from .recorder import Recorder, RecorderAttachable



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from ._base import (
    AlgoBase,
    AlgoConfigBase,
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    Controller,
    ExtendedProcessorBase,
    GenericProcessorBase,
    ProcessorBase,
)
from ._utils import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    PERCEIVED_WAVELENGTH,
    RLG_PER_HWAAS_MAP,
    PeakSortingMethod,
    ReflectorShape,
    _convert_amplitude_to_strength,
    _convert_multiple_amplitudes_to_strengths,
    calc_processing_gain,
    calculate_loopback_peak_location,
    double_buffering_frame_filter,
    exponential_smoothing_coefficient,
    find_peaks,
    get_approx_fft_vels,
    get_distance_filter_coeffs,
    get_distance_filter_edge_margin,
    get_distance_offset,
    get_distances_m,
    get_temperature_adjustment_factors,
    interpolate_peaks,
    select_prf,
    select_prf_m,
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import contextlib
from copy import copy
from typing import List, Optional

from PySide6.QtWidgets import (
    QButtonGroup,
    QFrame,
    QHBoxLayout,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

import pyqtgraph as pg

from acconeer.exptool.app.new.ui.utils import LayoutWrapper, LeftAlignDecorator


class TabPGWidget(QFrame):
    """
    Custom tab widget to handle GraphicsLayoutWidget.
    Since QTabWidget and GraphicsLayoutWidget caused GUI to freeze.
    """

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)

        self._button_group = QButtonGroup()
        self._button_layout = QHBoxLayout()
        self._button_layout.setContentsMargins(0, 0, 0, 0)

        self._main_layout = QVBoxLayout()
        button_wrapper = LeftAlignDecorator(LayoutWrapper(self._button_layout))
        self._main_layout.addWidget(button_wrapper)

        self._plot_widgets: List[pg.GrachicsLayoutWidget] = []

        self.setLayout(self._main_layout)

    def newPlotWidget(self, title: str) -> pg.GraphicsLayoutWidget:
        plot_widget = pg.GraphicsLayoutWidget()
        self.newTab(plot_widget, title)
        return plot_widget

    def newTab(self, widget: QWidget, title: str) -> None:
        new_tab_id = len(self._button_group.buttons())
        widget.setVisible(new_tab_id == 0)
        self._plot_widgets.append(widget)
        self._main_layout.addWidget(widget)

        button = QPushButton(title, self)
        button.setCheckable(True)
        button.setChecked(new_tab_id == 0)
        button.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self._button_group.addButton(button, id=len(self._button_group.buttons()))
        self._button_group.idToggled.connect(
            lambda button_id, checked: widget.setVisible(button_id == new_tab_id and checked)
        )
        self._button_layout.addWidget(button)

    def clear(self) -> None:
        if len(self._button_group.buttons()) > 0:
            with contextlib.suppress(RuntimeError, RuntimeWarning):
                self._button_group.idToggled.disconnect()

        for plot_widget in self._plot_widgets:
            self._main_layout.removeWidget(plot_widget)
            plot_widget.deleteLater()

        self._plot_widgets = []

        for button in copy(self._button_group.buttons()):
            self._button_group.removeButton(button)
            self._button_layout.removeWidget(button)
            button.deleteLater()
